from mewpy.model.smoment import SMomentModel
from mewpy.simulation import get_simulator
from mewpy.problems import GeckoOUProblem
from mewpy.utils.utilities import population_to_csv
import os


def test1(compound = 'R_EX_tyr__L_e'):
    """
    AutoPACMEN, "Automatic construction of metabolic models with enzyme constraints" (https://doi.org/10.1186/s12859-019-3329-9), is able to construct GECKO like models.
    This example optimizes the production of a compound using the E. coli autoPACMEN model where enzymes are defined as pseudo reactants.
    
    """
    DIR = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(DIR, '../../../examples/models/autopacmen/')
    DATA_FILE = os.path.join(PATH, "iJO1366_2019_06_25_GECKO.xml")

    model = SMomentModel(DATA_FILE, enzyme_reaction_prefix='R__TG_ER_')

    print(model.proteins)

    sim = get_simulator(model)
    sim.set_objective("R_BIOMASS_Ec_iJO1366_core_53p95M")
    solution = sim.simulate()
    print(solution)
    print('Wild type tyrosine production :',solution.fluxes['R_EX_tyr__L_e'])
    
    from mewpy.optimization.evaluation import BPCY,WYIELD
    from mewpy.simulation import SimulationMethod  
    from collections import OrderedDict

    envcond = OrderedDict()
    envcond.update({'R_EX_glc__D_e': (-10.0, 100000.0)})


   
    # the evaluation (objective) functions
    evaluator_1 = BPCY("R_BIOMASS_Ec_iJO1366_core_53p95M", compound, method=SimulationMethod.lMOMA)
    evaluator_2 = WYIELD("R_BIOMASS_Ec_iJO1366_core_53p95M", compound)
    
    
    problem = GeckoOUProblem(model,
                              fevaluation=[evaluator_1, evaluator_2],
                              envcond=envcond,
                              candidate_max_size=6,
                              prot_prefix='R__TG_ER_')

    # A new instance of the EA optimizer
    from mewpy.optimization import EA
    ea = EA(problem, max_generations=500)
    # runs the optimization
    final_pop = ea.run()

    import mewpy.utils.utilities as utl
    from time import time
    
    millis = int(round(time() * 1000))
    filename = "sMOMEMT{}_OU_{}.csv".format(compound, millis)
    utl.population_to_csv(problem, final_pop, filename, simplify=False)
    
    


def test2(compoud = 'R_EX_tyr__L_e',filename = None):
    """
    AutoPACMEN, "Automatic construction of metabolic models with enzyme constraints" (https://doi.org/10.1186/s12859-019-3329-9), is able to construct GECKO like models.
    This example optimizes the production of a compound using the E. coli autoPACMEN model where enzymes are defined as pseudo reactants.
    The model defines a linear constraint over the protein pool as reactant, by adding the protein pool as a metabolite in the stochiometric matrix.
    Therefore, the model may be treated as a regular GSM. 

    """
    DIR = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(DIR, '../../../examples/models/autopacmen/')
    DATA_FILE = os.path.join(PATH, "iJO1366_sMOMENT_2019_06_25.xml")
    from reframed.io.sbml import load_cbmodel
    model = load_cbmodel(DATA_FILE)
    
    #from cobra.io import read_sbml_model
    #model = read_sbml_model(DATA_FILE)
    # increases the enzyme pool upper bound
    #model.reactions.get_by_id('R_ER_pool_TG_').upper_bound = 1

    sim = get_simulator(model)
    s = sim.simulate()
    print(s)
    #print(s.fluxes)
    print('Wildtype tyrosine production :',s.fluxes['R_EX_tyr__L_e'])
    print('Pool:',s.fluxes['R_ER_pool_TG_'])
    
    fva = sim.FVA(reactions=['R_ER_pool_TG_'])
    print(fva)
    
    #for r_id, reaction in model.reactions.items():
    #    if reaction.gpr:
    #        print(str(reaction.gpr))

    #for rx in model.reactions:
    #        reaction = model.reactions.get_by_id(rx.id)
    #        if reaction.gene_reaction_rule:
    #           print(str(reaction.gene_reaction_rule))
        

    # implements a knockout over genes that encode enzymes
    BIOMASS_ID = 'R_BIOMASS_Ec_iJO1366_WT_53p95M'
    PRODUCT_ID = compoud
    from mewpy.optimization.evaluation import BPCY,WYIELD,TargetFlux
    from mewpy.simulation import SimulationMethod  
    evaluator_1 = BPCY(BIOMASS_ID, PRODUCT_ID, method=SimulationMethod.lMOMA)
    evaluator_2 = WYIELD(BIOMASS_ID, PRODUCT_ID)
    # evaluator_3 = TargetFlux(PRODUCT_ID)
    from mewpy.problems.genes import GOUProblem
    problem = GOUProblem(model, fevaluation=[
                         evaluator_1, evaluator_2], max_candidate_size =30)
    print(problem.target_list)
    from mewpy.optimization import EA
    ea = EA(problem, max_generations=20, mp=True)
    final_pop = ea.run()
    if filename:
        print("Simplifying and saving solutions to file")
        population_to_csv(problem, final_pop, filename, simplify=False)



if  __name__ == "__main__":
    for i in range(1):
        test1()



from collections import OrderedDict
from libsbml import AssignmentRule,SBMLReader
from reframed.core.model import Model
import re
import os 
import warnings



def load_sbml(filename):
    """ Loads an SBML file.
    
    :param filename: SBML file path, str.
    :returns: SBMLModel
            
    """

    if not os.path.exists(filename):
        raise IOError("Model file was not found")

    reader = SBMLReader()
    document = reader.readSBML(str(filename))
    sbml_model = document.getModel()

    if sbml_model is None:
        document.printErrors()
        raise IOError(f'Failed to load model {filename}.')

    return sbml_model


def load_ODEModel(filename, map = None):
    sbml_model = load_sbml(filename)
    from reframed.io.sbml import load_compartments,load_metabolites,load_reactions
    ode_model = ODEModel(sbml_model.getId())
    load_compartments(sbml_model, ode_model)
    load_metabolites(sbml_model, ode_model)
    ## adds constants and boundaries to metabolites
    for species in sbml_model.getListOfSpecies():
        m = ode_model.metabolites[species.getId()]
        if not hasattr(m,"constant"):
            m.constant = species.getConstant()
        if not hasattr(m,"boundary"):
            m.boundary = species.getBoundaryCondition()
    load_reactions(sbml_model, ode_model)
    _load_concentrations(sbml_model, ode_model)
    _load_global_parameters(sbml_model, ode_model)
    _load_local_parameters(sbml_model, ode_model)
    _load_ratelaws(sbml_model, ode_model)
    _load_assignment_rules(sbml_model, ode_model)
    # parse rates, rules and xdot expressions
    ode_model._set_parsed_attr()
    if isinstance(map,str):
        print("MAP =",map)
        aux = OrderedDict([(rId, re.findall(map, ratelaw)) for rId, ratelaw in ode_model.ratelaws.items()])
        #aux = OrderedDict([(rId ,  [rId+"_"+x for x in re.findall("(rmax\w*)", ratelaw)]) for rId, ratelaw in model.ratelaws.items()])#CHASSAGNOLE
        ode_model.reacParamsFactors = OrderedDict([(rId , params) for rId, params in aux.items() if len(params) > 0])
    else:
        ode_model.set_reactions_parameters_factors(map)

        
    return ode_model




def _load_concentrations(sbml_model, odemodel):
    for species in sbml_model.getListOfSpecies():
        odemodel.set_concentration(species.getId(), species.getInitialConcentration())


def _load_global_parameters(sbml_model, odemodel):
    for parameter in sbml_model.getListOfParameters():
            odemodel.set_global_parameter(parameter.getId(), parameter.getValue(), parameter.getConstant())


def _load_local_parameters(sbml_model, odemodel):
    for reaction in sbml_model.getListOfReactions():
        for parameter in reaction.getKineticLaw().getListOfParameters():
            odemodel.set_local_parameter(reaction.getId(), parameter.getId(), parameter.getValue())


def _load_ratelaws(sbml_model, odemodel):
    for reaction in sbml_model.getListOfReactions():
        odemodel.set_ratelaw(reaction.getId(), reaction.getKineticLaw().getFormula())


def _load_assignment_rules(sbml_model, odemodel):
    for rule in sbml_model.getListOfRules():
        if isinstance(rule, AssignmentRule):
            odemodel.set_assignment_rule(rule.getVariable(), rule.getFormula())





class ODEModel(Model):

    def __init__(self, model_id):
        """ ODE Model.
        
        :param model: a REFRAMED SBModel or COBRApy Model 
            
        """
        super(ODEModel,self).__init__(model_id)
        self.concentrations = OrderedDict()
        self.constant_params = OrderedDict()
        self.variable_params = OrderedDict()
        self.local_params = OrderedDict()
        self.ratelaws = OrderedDict()
        self.assignment_rules = OrderedDict()
        self._func_str = None
        self._constants = None
        self.reacParamsFactors = None  
        self.parsedRates = None
        self.parsedRules = None
        self.parsedXdot = None
        


    def get_reactions(self):
        return self.reactions

    def _clear_temp(self):
        self.update()
        self._func_str = None

    def add_reaction(self, reaction, replace=True, ratelaw=''):
        super(ODEModel,self).add_reaction(reaction,replace)
        self.ratelaws[reaction.id] = ratelaw
        self.local_params[reaction.id] = OrderedDict()

    def set_concentration(self, m_id, concentration):
        if m_id in self.metabolites:
            self.concentrations[m_id] = concentration
        else:
            warnings.warn("No such metabolite '{}'".format(m_id), RuntimeWarning)

    def set_ratelaw(self, r_id, ratelaw):
        if r_id in self.reactions:
            self.ratelaws[r_id] = ratelaw
        else:
            warnings.warn("No such reaction '{}'".format(r_id), RuntimeWarning)

    def set_assignment_rule(self, p_id, rule):
        if p_id in self.variable_params or p_id in self.metabolites:
            self.assignment_rules[p_id] = rule
        else:
            warnings.warn("No such variable parameter '{}'".format(p_id), RuntimeWarning)

    def set_global_parameter(self, key, value, constant=True):
        if constant:
            self.constant_params[key] = value
        else:
            self.variable_params[key] = value

    def set_local_parameter(self, r_id, p_id, value):
        if r_id in self.reactions:
            if r_id not in self.local_params.keys():
                self.local_params[r_id] = OrderedDict()
            self.local_params[r_id][p_id] = value
        else:
            warnings.warn("No such reaction '{}'".format(r_id), RuntimeWarning)

    def remove_reactions(self, id_list):
        self.remove_reactions(id_list)
        for r_id in id_list:
            del self.ratelaws[r_id]
            del self.local_params[r_id]
            

    def merge_constants(self):
        constants = OrderedDict()

        for c_id, comp in self.compartments.items():
            constants[c_id] = comp.size

        constants.update(self.constant_params)

        for r_id, params in self.local_params.items():
            for p_id, value in params.items():
                full_id = '{}_{}'.format(r_id, p_id)
                constants[full_id] = value

        self._constants = constants
        return constants

    def get_parameters(self, exclude_compartments=False):
        if not self._constants:
            self.merge_constants()

        parameters = self._constants.copy()

        if exclude_compartments:
            for c_id in self.compartments:
                del parameters[c_id]

        return parameters


    def print_balance(self, m_id):
        c_id = self.metabolites[m_id].compartment
        table = self.metabolite_reaction_lookup()
        terms = ["{:+g} * r['{}']".format(coeff, r_id) for r_id, coeff in table[m_id].items()]
        if len(terms)==0 or (self.metabolites[m_id].constant and self.metabolites[m_id].boundary):
            expr= "0"
        else:
            expr = "1/p['{}'] * ({})".format(c_id, ' '.join(terms))
        return expr


    def parse_rate(self, r_id, rate):

        symbols = '()+*-/,'
        rate = ' ' + rate + ' '
        for symbol in symbols:
            rate = rate.replace(symbol, ' ' + symbol + ' ')

        for i, m_id in enumerate(self.metabolites):
            rate = rate.replace(' ' + m_id + ' ', ' x[{}] '.format(i))

        for c_id in self.compartments:
            rate = rate.replace(' ' + c_id + ' ', " p['{}'] ".format(c_id))

        for p_id in self.constant_params:
            if p_id not in self.local_params[r_id]:
                rate = rate.replace(' ' + p_id + ' ', " p['{}'] ".format(p_id))

        for p_id in self.variable_params:
            if p_id not in self.local_params[r_id]:
                rate = rate.replace(' ' + p_id + ' ', " v['{}'] ".format(p_id))

        for p_id in self.local_params[r_id]:
            rate = rate.replace(' ' + p_id + ' ', " p['{}_{}']".format(r_id, p_id))

        return rate

    def parse_rule(self, rule, parsed_rates):

        symbols = '()+*-/,'
        rule = ' ' + rule + ' '
        for symbol in symbols:
            rule = rule.replace(symbol, ' ' + symbol + ' ')

        for i, m_id in enumerate(self.metabolites):
            rule = rule.replace(' ' + m_id + ' ', ' x[{}] '.format(i))

        for c_id in self.compartments:
            rule = rule.replace(' ' + c_id + ' ', " p['{}'] ".format(c_id))

        for p_id in self.constant_params:
            rule = rule.replace(' ' + p_id + ' ', " p['{}'] ".format(p_id))

        for p_id in self.variable_params:
            rule = rule.replace(' ' + p_id + ' ', " v['{}'] ".format(p_id))

        for r_id in self.reactions:
           rule = rule.replace(' ' + r_id + ' ', '({})'.format(parsed_rates[r_id]))

        return rule

    def _set_parsed_attr(self):
        self.parsedRates = {rId: self.parse_rate(rId, ratelaw)
                             for rId, ratelaw in self.ratelaws.items()}
        
        aux = {pId: self.parse_rule(rule, self.parsedRates)
               for pId, rule in self.assignment_rules.items()}
        
        trees = [_build_tree_rules(vId, aux) for vId in aux.keys()]
        order = _get_oder_rules(trees)
        
        self.parsedRules = OrderedDict([(id, aux[id]) for id in order])
        self.parsedXdot = {mId: self.print_balance(mId) for mId in self.metabolites}



    def build_ode(self, factors):
        """
        Builds de ODE model
        
            
        :param factors: (dict) The key is the parameter identifier and the value is the level of change values between 0 and 1 represent a under expression, above 1 a over expression and 0 to represent the knockouts.
        :returns: Returns  a string with the ode system.
        
        """

        # factors: ["vmax1": 0, "vmax2"=2, "ENZYME_ID":0]
        # divide vmax parameters from enzymes expression levels
        factorsEnz = OrderedDict([(k, v) for k, v in factors.items() if k in self.metabolites])
        factorsParam = OrderedDict([(k, v) for k, v in factors.items() if k not in factorsEnz.keys()])

        ruleExprs = ["    v['{}'] = {}".format(pId, self.parsedRules[pId])
                     for pId in self.parsedRules.keys()]

        rateExprs = []
        for rId in self.reactions:
            newExp = self.parsedRates[rId]
            if rId in self.reacParamsFactors.keys():
                toModify = set(factorsParam.keys()).intersection(self.reacParamsFactors[rId])
                if len(toModify) > 0:
                    for elem in toModify:
                        newExp = re.sub(r"([pv]\['" + elem + "'\])", str(factorsParam[elem]) + r" * \1", newExp)
            rateExprs.append("    r['{}'] = {}".format(rId, newExp))

        balances = []
        for m_id in self.metabolites:
            exp = self.parsedXdot[m_id]
            if m_id in factorsEnz.keys():
                if factors[m_id] == 0:
                    newExp = "0"
                else:
                    newExp = re.sub(r"\+\s*(\d)", r"+ \1 * " + str(factorsEnz[m_id]), exp)
                balances.append(' ' * 8 + newExp)
            else:
                balances.append(' ' * 8 + exp)

        func_str = 'def ode_func(t, x, r, p, v):\n\n' + \
                   '\n'.join(ruleExprs) + '\n\n' + \
                   '\n'.join(rateExprs) + '\n\n' + \
                   '    dxdt = [\n' + \
                   ',\n'.join(balances) + '\n' + \
                   '    ]\n\n' + \
                   '    return dxdt\n'
        return func_str


    def get_ode(self, r_dict=None, params=None, factors=None):
        """
        Build the ODE system.
        
        :param r_dict: (dict) This variable is used to store the reaction rates.
        :param params: (dict) Parameters and the new values used to replace the original parameters present in the SBML model.
        :param factors: (dict) The key is the parameter identifier and the value is the level of change values between 0 and 1 represent a under expression, above 1 a over expression and 0 to represent the knockouts.
        :returns: A function used to solve the ODE system.
            
        """

        p = self.merge_constants()
        v = self.variable_params.copy()

        if r_dict is not None:
            r = r_dict
        else:
            r = {}

        if params:
            p.update(params)

        exec(self.build_ode(factors), globals())
        ode_func = eval('ode_func')
        #print "Parameters"
        #print p
        #print v
        #print "-----"
        #print (self.build_ode(factors))
        f = lambda t, x: ode_func(t, x, r, p, v)
        return f

    def set_reactions_parameters_factors(self, map):
        """
        Set a new map with the parameters that can be changed for each reaction.
        
        :param map: (dict) The keys is the reaction identifier and the value a list of parameters which can be used to simulate modifications( KO, under/ over expression)
        
        """

        self.reacParamsFactors = OrderedDict(map) if map else OrderedDict()


    def __getstate__(self):
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)




# auxiliar functions to set the assignment rules by the correct order in the ODE system
def _build_tree_rules(parent, rules):
    regexp = "v\[\'(.*?)\'\]"
    children = re.findall(regexp, rules[parent])
    if len(children) == 0:
        return Tree(parent, None)
    else:
        childrenTrees = [_build_tree_rules(child, rules) for child in children]
        return Tree(parent, childrenTrees)


def _get_oder_rules(trees):
    res = []
    for tree in trees:
        new_elems = _get_order_nodes(tree)
        [res.append(item) for item in new_elems if item not in res]
    return res


def _get_order_nodes(tree):
    res = [tree.name]
    if len(tree.children) > 0:
        for child in tree.children:
            res = _get_order_nodes(child) + res
    return res

class Tree(object):
    "Generic tree node."
    def __init__(self, name='root', children=None):
        self.name = name
        self.children = []
        if children is not None:
            for child in children:
                self.add_child(child)

    def add_child(self, node):
       # assert isinstance(node, MyTree)
        self.children.append(node)

def get_order_nodes(tree):
    if tree.children is None:
        return [tree.name]
    else:
        res = []
        for child in tree.children:
            res = res + get_order_nodes(child)
        return res


""" Inspyred Operators
"""

from inspyred.ec.variators.crossovers import crossover
from inspyred.ec.variators.mutators import mutator
import copy


@mutator
def shrink_mutation(random, candidate, args):
    """Returns the mutant produced by shrink mutation on the candidate.
    If a candidate solution has length of 1, this function leaves it unchanged.

    :param random: the random number generator object. Must implement random().
    :param candidate: the candidate solution.
    :parm args: a dictionary of keyword arguments.
    :returns: A set of new candidate

    Notes:
        Optional keyword arguments in args:
        - *mutation_rate* -- the rate at which mutation is performed (default 0.1)
    """
    mutRate = args.setdefault("gs_mutation_rate", 0.1)
    minSize = args["candidate_min_size"]
    if random.random() > mutRate or len(candidate) <= minSize:
        return candidate
    index = random.randint(0, len(candidate) - 1)
    mutantL = list(candidate)
    del mutantL[index]
    mutant = set(mutantL)
    return mutant


@crossover
def uniform_crossover_KO(random, mom, dad, args):
    """Return the offspring of the uniform crossover on the candidate. Based on two candidates (parents) build 2 children:
    - elements present in both parents will be present in both children;
    - both children have at least one element;
    - elements present in only one parent have equal probability to be present in child 1 or child 2 (after each child
      has at least one element).

    :param random: the random number generator object. Must implement random().
    :param mom: the first parent candidate
    :param dad: the second parent candidate
    :param args: a dictionary of keyword arguments.
    :returns: Return the offspring of the candidates given as argument.

    Notes:
        Optional keyword arguments in args:
        - *crossover_rate* -- the rate at which crossover is performed (default 1.0)
    """
    crossRate = args.setdefault("crossover_rate", 1.0)
    children = []
    if random.random() > crossRate or (len(mom) == 1 and len(dad) == 1):
        children.append(mom)
        children.append(dad)
        return children

    maxSize = args["candidate_max_size"]
    intersection = mom & dad
    otherElems = list((mom | dad).difference(intersection))
    child1 = copy.copy(intersection)
    child2 = copy.copy(intersection)

    while len(otherElems) > 0:
        elemPosition = random.randint(
            0, len(otherElems) - 1) if len(otherElems) > 1 else 0
        if len(child1) == maxSize or len(child2) == 0:
            child2.add(otherElems[elemPosition])
        elif len(child2) == maxSize or len(child1) == 0:
            child1.add(otherElems[elemPosition])
        else:
            r = random.random()
            if r <= 0.5:
                child1.add(otherElems[elemPosition])
            else:
                child2.add(otherElems[elemPosition])
        otherElems.pop(elemPosition)
    children.append(child1)
    children.append(child2)
    return children


@crossover
def uniform_crossover_OU(random, mom, dad, args):
    """Return the offspring of the uniform crossover on the candidate. Based on two candidates (parents) build 2 children:
    - elements present in both parents will be present in both children;
    - both children have at least one element;
    - elements present in only one parent have equal probability to be present in child 1 or child 2 (after each child
    has at least one element).

    :param random: the random number generator object. Must implement random().
    :param mom: the first parent candidate
    :param dad: the second parent candidate
    :param args: a dictionary of keyword arguments.
    :returns: Return the offspring of the candidates given as argument.

    Notes:
        Optional keyword arguments in args:
        - *crossover_rate* -- the rate at which crossover is performed (default 1.0)
    """
    crossRate = args.setdefault("crossover_rate", 1.0)
    children = []
    if random.random() > crossRate or (len(mom) == 1 and len(dad) == 1):
        children.append(copy.copy(mom))
        children.append(copy.copy(dad))
        return children

    maxSize = args["candidate_max_size"]
    # common idx (reactions)
    intersection = list({idx for (idx, idy) in mom} &
                        {idx for (idx, idy) in dad})
    c_mom = {idx: idy for (idx, idy) in mom if idx in intersection}
    c_dad = {idx: idy for (idx, idy) in dad if idx in intersection}
    rest = [(idx, idy) for (idx, idy) in mom if idx not in intersection] + \
        [(idx, idy) for (idx, idy) in dad if idx not in intersection]
    child1 = []
    child2 = []

    for i in range(len(intersection)):
        idx = intersection[i]
        if random.random() < 0.5:
            child1.append((idx, c_mom[idx]))
            child2.append((idx, c_dad[idx]))
        else:
            child2.append((idx, c_mom[idx]))
            child1.append((idx, c_dad[idx]))

    for i in range(len(rest)):
        if random.random() < 0.5 and len(child1) < maxSize:
            child1.append(rest[i])
        elif len(child2) < maxSize:
            child2.append(rest[i])
        else:
            child1.append(rest[i])

    if len(child1) == 0:
        i = random.randint(0, len(child2)-1)
        p = child2[i]
        child1.append(p)
        child2.remove(p)

    if len(child2) == 0:
        i = random.randint(0, len(child1)-1)
        p = child1[i]
        child2.append(p)
        child1.remove(p)

    children.append(set(child1))
    children.append(set(child2))
    return children


@mutator
def grow_mutation_KO(random, candidate, args):
    """Returns the mutant produced by a grow mutation on the candidate (when the representation is a set of integers).
    If a candidate solution has the maximum size candidate allowed, this function leaves it unchanged.

    :param random: the random number generator object.
    :param candidate: the candidate solution.
    :param args: a dictionary of keyword arguments.
    :returns: A set with a new candidate.

    Notes:

    Optional keyword arguments in args:
    - *mutation_rate* -- the rate at which mutation is performed (default 0.1)
    """
    bounder = args["_ec"].bounder
    mutRate = args.setdefault("gs_mutation_rate", 0.1)
    if random.random() > mutRate:
        return candidate
    maxSize = args["candidate_max_size"]
    mutant = copy.copy(candidate)
    if len(mutant) < maxSize:
        newElem = random.randint(bounder.lower_bound, bounder.upper_bound)
        while newElem in mutant:
            newElem = random.randint(bounder.lower_bound, bounder.upper_bound)
        mutant.add(newElem)
    return mutant


@mutator
def grow_mutation_OU(random, candidate, args):
    """Returns the mutant produced by a grow mutation on the candidate (when the representation is a set of integers).
    If a candidate solution has the maximum size candidate allowed, this function leaves it unchanged.

    :param random: the random number generator object.
    :param candidate: the candidate solution.
    :param args: a dictionary of keyword arguments.
    :returns: A set with a new candidate.

    Notes:

    Optional keyword arguments in args:
    - *mutation_rate* -- the rate at which mutation is performed (default 0.1)

    """
    bounder = args["_ec"].bounder
    mutRate = args.setdefault("gs_mutation_rate", 0.1)
    if random.random() > mutRate:
        return candidate
    maxSize = args["candidate_max_size"]
    mutant = copy.copy(candidate)
    if len(mutant) < maxSize:
        idx = random.randint(bounder.lower_bound[0], bounder.upper_bound[0])
        idxs = [a for (a, b) in mutant]
        while idx in idxs:
            idx = random.randint(
                bounder.lower_bound[0], bounder.upper_bound[0])
        lv = random.randint(bounder.lower_bound[1], bounder.upper_bound[1])
        mutant.add((idx, lv))
    return mutant


@mutator
def single_mutation_KO(random, candidate, args):
    """Returns the mutant produced by a single mutation on the candidate (when the representation is a set of integers).
    The candidate size is maintained.

    :param random: the random number generator object.
    :param candidate: the candidate solution.
    :param args: a dictionary of keyword arguments.
    :returns: A set with a new candidate.

    Optional keyword arguments in args:

    - *mutation_rate* -- the rate at which mutation is performed (default 0.1)
    """

    bounder = args["_ec"].bounder
    mutRate = args.setdefault("mutation_rate", 0.1)
    if random.random() > mutRate:
        return candidate
    mutant = copy.copy(candidate)
    index = random.randint(0, len(mutant) - 1) if len(mutant) > 1 else 0
    newElem = random.randint(bounder.lower_bound, bounder.upper_bound)
    while newElem in mutant:
        newElem = random.randint(bounder.lower_bound, bounder.upper_bound)
    mutantL = list(mutant)
    mutantL[index] = newElem
    mutant = set(mutantL)
    return mutant


@mutator
def single_mutation_OU(random, candidate, args):
    """Returns the mutant produced by one mutation on the candidate (when the representation is a set of (int,int)).
    The candidate size is maintained.

    :param random: the random number generator object.
    :param candidate: the candidate solution.
    :param args: a dictionary of keyword arguments.
    :returns: A set with a new candidate.

    Optional keyword arguments in args:

    - *mutation_rate* -- the rate at which mutation is performed (default 0.1)
    """

    bounder = args["_ec"].bounder
    mutRate = args.setdefault("mutation_rate", 0.1)
    if random.random() > mutRate:
        return candidate

    mutant = copy.copy(candidate)
    index = random.randint(0, len(mutant) - 1) if len(mutant) > 1 else 0
    # the first idx has a 50% chance of beeing mutated
    # the second always mutates
    ml = [i for (i, j) in mutant]
    mutantL = list(mutant)
    idx, idy = mutantL[index]
    is_mutate_idx = False
    if random.random() > 0.5:
        idx = random.randint(bounder.lower_bound[0], bounder.upper_bound[0])
        while idx in ml:
            idx = random.randint(
                bounder.lower_bound[0], bounder.upper_bound[0])
        is_mutate_idx = True
    lv = random.randint(bounder.lower_bound[1], bounder.upper_bound[1])
    while not is_mutate_idx and lv == idy:
        lv = random.randint(bounder.lower_bound[1], bounder.upper_bound[1])
    mutantL[index] = (idx, lv)
    mutant = set(mutantL)
    return mutant


@mutator
def single_mutation_OU_level(random, candidate, args):
    """Returns the mutant produced by one mutation on the candidate (when the representation is a set of (int,int)).
    The candidate size is maintained.

    :param random: the random number generator object.
    :param candidate: the candidate solution.
    :param args: a dictionary of keyword arguments.
    :returns: A set with a new candidate.


    Optional keyword arguments in args:

    - *mutation_rate* -- the rate at which mutation is performed (default 0.1)

    """

    bounder = args["_ec"].bounder
    mutRate = args.setdefault("mutation_rate", 0.1)
    if random.random() > mutRate:
        return candidate
    mutant = copy.copy(candidate)
    index = random.randint(0, len(mutant) - 1) if len(mutant) > 1 else 0
    # the first idx has a 50% chance of beeing mutated
    # the second always mutates
    mutantL = list(mutant)
    idx, idy = mutantL[index]
    lv = random.randint(bounder.lower_bound[1], bounder.upper_bound[1])
    while lv == idy:
        lv = random.randint(bounder.lower_bound[1], bounder.upper_bound[1])
    mutantL[index] = (idx, lv)
    mutant = set(mutantL)
    return mutant



"""
Implements a crossmodel simulation of solutions
"""
import pandas as pd

map_problem_notation = {"GeckoRKOProblem": "Protein",
                        "GeckoROUProblem": "Protein",
                        "GKOProblem": "Systematic",
                        "GOUProblem": "Systematic",
                        "OptRamProblem": "Systematic"
                        }


class NotationTranslator:
    def __init__(self, database, from_notation, to_notation, sep=';'):
        if isinstance(database, pd.DataFrame):
            self.db = database
        elif isinstance(database, str):
            self.db = pd.read_csv(database, sep=sep)
        self.from_notation = from_notation
        self.to_notation = to_notation

    def columns(self):
        return self.db.columns.tolist()

    def translate(self, value):
        """
        Translate a single value
        """
        if self.from_notation == self.to_notation:
            return value

        la = self.db.loc[self.db[self.from_notation] == value][self.to_notation].tolist()
        if la and len(la) > 0:
            return la[0]
        else:
            raise ValueError(f'Value {value} not found.')

    def translate_representation(self, representation, source_prefix="", destination_prefix=""):
        """
        Translates constraints.
        constraints are defined as a dictionary (OU) or a list (KO)
        """

        p = len(source_prefix)
        if isinstance(representation, list):
            return [destination_prefix+self.translate(value[p:]) for value in representation]
        else:
            return {destination_prefix+self.translate(value[p:]): level for value, level in representation.items()}

    def get_list(self, name):
        return self.db[name].tolist()



"""ODE Solver
"""
from enum import Enum


class SolverConfigurations:
    ABSOLUTE_TOL = 1e-9
    RELATIVE_TOL = 1e-6
    N_STEPS = 10000


class SolverMethod(Enum):
    LSODA = 1
    LSODAR = 2
    LSODE = 3
    HEUN = 4
    EULER = 5
    RK4 = 6
    DORMAN_PRINCE = 7
    RKFehlberg = 8
    Dopri5 = 9
    Dop853 = 10
    Vode = 11
    Radau5 = 12
    AdamsBashforth2 = 13
    AdamsBashMoulton2 = 14


class KineticConfigurations:
    STEADY_STATE_TIME = 1e9
    SOLVER = "odespy"
    SOLVER_METHOD = SolverMethod.HEUN  # ode solver method used in the phenotype simulation
    SOLVER_TIMEOUT = 6000  # maximum time allowed by simulation


class ODEStatus(Enum):
    ERROR = 0
    OPTIMAL = 1


class ODESpySolver:
    """
    ODE solver method implemented on odespy package.
    """

    def __init__(self, solverMethod):
        self.solverMethod = solverMethod

    def get_solver(self, func):
        """
        Returns the solver method from odespy package.

        :param func: function with ODE system.
        :return: an instance of odeSolver

        """
        try:
            import odespy
        except Exception:
            raise RuntimeError("ODEspy package is required")

        if self.solverMethod is SolverMethod.LSODA:
            solver = odespy.Lsoda(func)
        elif self.solverMethod is SolverMethod.LSODAR:
            solver = odespy.Lsodar(func)
        elif self.solverMethod is SolverMethod.LSODE:
            solver = odespy.Lsode(func)
        elif self.solverMethod is SolverMethod.HEUN:
            solver = odespy.Heun(func)
        elif self.solverMethod is SolverMethod.EULER:
            solver = odespy.Euler(func)
        elif self.solverMethod is SolverMethod.RK4:
            solver = odespy.RK4(func)
        elif self.solverMethod is SolverMethod.DORMAN_PRINCE:
            solver = odespy.DormandPrince(func)
        elif self.solverMethod is SolverMethod.RKFehlberg:
            solver = odespy.RKFehlberg(func)
        elif self.solverMethod is SolverMethod.Dopri5:
            solver = odespy.Dopri5(func)
        elif self.solverMethod is SolverMethod.Dop853:
            solver = odespy.Dop853(func)
        elif self.solverMethod is SolverMethod.Vode:
            solver = odespy.Vode(func)
        elif self.solverMethod is SolverMethod.AdamsBashforth2:
            solver = odespy.AdamsBashforth2(func, method='bdf')
        elif self.solverMethod is SolverMethod.Radau5:
            solver = odespy.Radau5(func)
        elif self.solverMethod is SolverMethod.AdamsBashMoulton2:
            solver = odespy.AdamsBashMoulton2(func)

        # update default parameters
        solver.nsteps = SolverConfigurations.N_STEPS
        solver.atol = SolverConfigurations.ABSOLUTE_TOL
        solver.rtol = SolverConfigurations.RELATIVE_TOL

        return solver

    def __getstate__(self):
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)



from abc import ABC, abstractmethod
from mewpy.utils.constants import EAConstants

import copy
# pathos
try:
    import pathos.multiprocesssing
    from pathos.multiprocessing import Pool
except ImportError:
    import multiprocessing
    from multiprocessing.pool import Pool
# dask
try:
    import dask
except ImportError:
    pass
# pyspark
try:
    from pyspark import SparkConf, SparkContext
except ImportError:
    pass


def cpu_count():
    """ The number of cpus
        Return EAConstants.NUM_CPUS if it is set to a positive number
        otherwise return half of the available cpus.
    """
    if EAConstants.NUM_CPUS > 0:
        return EAConstants.NUM_CPUS
    else:
        try:
            return multiprocessing.cpu_count()//2
        except (ImportError, NotImplementedError):
            return 1


class Evaluator(ABC):

    @abstractmethod
    def evaluate(self, candidates, args):
        raise NotImplementedError


class MultiProcessorEvaluator(Evaluator):

    def __init__(self, evaluator, mp_num_cpus):
        """
        Evaluate the candidates in parallel using ``multiprocessing``.
        """
        self.pool = Pool(mp_num_cpus)
        self.evaluator = evaluator

    def evaluate(self, candidates, args):
        """
        Values in args will be ignored and not passed to the evaluator to avoid unnecessary pickling in inspyred.
        """
        results = self.pool.map(self.evaluator, candidates)
        return results


class DaskEvaluator(Evaluator):

    def __init__(self, evaluator, mp_num_cpus, scheduler='processes'):
        self.evaluator = evaluator
        self.scheduler = scheduler

    def evaluate(self, candidates, args):
        with dask.config.set(scheduler=self.scheduler):
            return list(dask.compute(*[dask.delayed(self.evaluator)(c) for c in candidates]))


class SparkEvaluator(Evaluator):
    def __init__(self, evaluator, mp_num_cpus):
        self.evaluator = evaluator
        self.spark_conf = SparkConf().setAppName(
            "mewpy").setMaster(f"local[{mp_num_cpus}]")
        self.spark_context = SparkContext(conf=self.spark_conf)

    def evaluate(self, candidates, args):
        solutions_to_evaluate = self.spark_context.parallelize(candidates)

        return solutions_to_evaluate.map(lambda s: self.evaluator(s)).collect()


# ray
try:
    import ray
except ImportError:
    pass
else:

    @ray.remote
    class RayActor:
        """
        Each actor (worker) has a solver instance to overcome the need to serialize solvers which may not be pickable.
        The solver is not reset before each evaluation.
        """

        def __init__(self, problem):
            self.problem = copy.deepcopy(problem)

        def evaluate_candidates(self, candidates):
            """Evaluates a sublist of candidates
            """
            return self.problem.evaluator(candidates, None)

    class RayEvaluator(Evaluator):

        def __init__(self, problem, number_of_actors):
            ray.init(ignore_reinit_error=True)
            self.actors = [RayActor.remote(problem)
                           for _ in range(number_of_actors)]
            self.number_of_actors = len(self.actors)
            print(f"Using {self.number_of_actors} workers.")

        def evaluate(self, candidates, args):
            """
            Divides the candidates into sublists to be evaluated by each actor
            """
            size = len(candidates) // self.number_of_actors
            if len(candidates) % self.number_of_actors != 0:
                size += 1
            sub_lists = [candidates[i:i+size]
                         for i in range(0, len(candidates), size)]
            values = []
            for i in range(self.number_of_actors):
                actor = self.actors[i]
                values.append(actor.evaluate_candidates.remote(sub_lists[i]))
            r = ray.get(values)
            result = []
            for x in r:
                for y in x:
                    result.append(y)
            return result



from .regulatory_variable import RegulatoryVariable
from .regulatory_interaction import RegulatoryInteraction
from .regulatory_model import RegulatoryModel
from .integrated_model import IntegratedModel, IntegratedRegulatoryInteraction, IntegratedRegulatoryVariable
from .RFBA import RFBAModel
from .SRFBA import SRFBAModel
from .optram import load_optram, OptRAMRegModel, OptRamProblem



"""
EA comparison
"""
from jmetal.algorithm.multiobjective.nsgaiii import NSGAIII, UniformReferenceDirectionFactory
from jmetal.algorithm.multiobjective.nsgaii import NSGAII
from jmetal.algorithm.multiobjective.spea2 import SPEA2
from jmetal.algorithm.multiobjective.ibea import IBEA
from jmetal.core.quality_indicator import GenerationalDistance, EpsilonIndicator, HyperVolume
from jmetal.lab.experiment import Experiment, Job, generate_summary_from_experiment
from jmetal.util.termination_criterion import StoppingByEvaluations
from jmetal.operator import BinaryTournamentSelection
from jmetal.util.comparator import Comparator, MultiComparator
from jmetal.util.evaluator import MultiprocessEvaluator



candidate_max_size = 6
max_evaluations = 100
N_CPU = 2


def configure_experiment(problems: dict, n_run: int):

    from mewpy.optimization.jmetal.operators import UniformCrossoverOU, GrowMutationOU, ShrinkMutation,SingleMutationOU, MutationContainer
    crossover = UniformCrossoverOU(0.5,max_size=candidate_max_size)
    mutators = []
    mutators.append(GrowMutationOU(1.0, max_size=candidate_max_size))
    mutators.append(ShrinkMutation(1.0, min_size=candidate_max_size))
    mutators.append(SingleMutationOU(1.0))
    mutation = MutationContainer(0.3, mutators=mutators)


    jobs = []
    

    for run in range(n_run):
        for problem_tag, problem in problems.items():
            jobs.append(
                Job(
                    algorithm=NSGAII(
                        problem=problem,
                        population_evaluator=MultiprocessEvaluator(N_CPU),
                        population_size=100,
                        offspring_population_size=100,
                        mutation= mutation,
                        crossover=crossover,
                        termination_criterion=StoppingByEvaluations(max_evaluations=max_evaluations)
                        ),
                    algorithm_tag='NSGAII',
                    problem_tag=problem_tag,
                    run=run,
                )
            )
            jobs.append(
                Job(
                    algorithm=SPEA2(
                        problem=problem,
                        population_evaluator=MultiprocessEvaluator(N_CPU),
                        population_size=100,
                        offspring_population_size=100,
                        mutation= mutation,
                        crossover=crossover,
                        termination_criterion=StoppingByEvaluations(max_evaluations=max_evaluations)
                        ),
                    algorithm_tag='SPEA2',
                    problem_tag=problem_tag,
                    run=run
                    )
            )
            jobs.append(
                Job(
                    algorithm=IBEA(
                        problem=problem,
                        population_evaluator=MultiprocessEvaluator(N_CPU),
                        kappa=1.,
                        population_size = 100, 
                        offspring_population_size = 100,
                        mutation=mutation,
                        crossover=crossover,
                        termination_criterion=StoppingByEvaluations(max_evaluations=max_evaluations)
                        ),
                    algorithm_tag='IBEA',
                    problem_tag=problem_tag,
                    run=run,
                )
            )
        

    return jobs


if __name__ == '__main__':

    # Run the study
    output_directory = 'data'

    
    # Configure the experiments
    from mewpy.model.gecko import GeckoModel 
    from collections import OrderedDict 
    compound =  'r_1913'
    model = GeckoModel('single-pool', biomass_reaction_id='r_2111')
    model.set_objective({'r_2111': 1.0, 'r_4041': 0.0})
    envcond = OrderedDict()

    from mewpy.optimization.evaluation import BPCY,WYIELD
    from mewpy.simulation import SimulationMethod
    # the evaluation (objective) functions
    evaluator_1 = BPCY("r_2111", compound, method=SimulationMethod.lMOMA)
    evaluator_2 = WYIELD("r_2111", compound)

    from mewpy.problems import GeckoOUProblem
    p = GeckoOUProblem(model,
                              fevaluation=[evaluator_1, evaluator_2],
                              candidate_max_size=candidate_max_size)

    from mewpy.optimization.jmetal.problem import JMetalOUProblem
    problem = JMetalOUProblem(p)
    problem.number_of_objectives = 2
    
    problems = {'gecko':problem}
    
    jobs = configure_experiment(problems=problems, n_run=1)

    
    experiment = Experiment(output_dir=output_directory, jobs=jobs)
    experiment.run()

    import numpy
    import os
    from mewpy.optimization.ea import non_dominated_population, Solution

    for prb in problems.keys():
        # Builds a pareto front approximation from the results
        population =[]
        for r, _, fl in os.walk(output_directory):
                for file in fl:
                    if 'FUN' in file and prb in r:
                        with open(os.path.join(r, file), 'r') as f:
                            line = f.readline()
                            while line:
                                tokens = line.split()
                                fitness = [float(x) for x in tokens]
                                s = Solution(None,fitness)
                                population.append(s)
                                line = f.readline()
            
        population = non_dominated_population(population, maximize=False, filter_duplicate=False)
        # save to file
        pf_file = os.path.dirname(output_directory).join(prb+".pf")
        with open(pf_file,'w') as f:
            for s in population:
                f.write("".join([str(v)+"\t" for v in s.fitness]))
                f.write("\n")

    # Generate summary file
    generate_summary_from_experiment(
        input_dir=output_directory,
        quality_indicators=[GenerationalDistance(reference_front = output_directory), 
                            EpsilonIndicator(reference_front = output_directory),
                            HyperVolume([0,0])
                            ] 
    )


from warnings import warn


class RegulatoryVariable(object):

    def __init__(self, identifier,
                 name=None,
                 expression_coef=None,
                 aliases=None,
                 regulatory_interactions=None,
                 targets=None,
                 regulators=None,
                 model=None,
                 is_regulator=False,
                 is_target=False):

        if not name:
            name = ''

        if not aliases:
            aliases = []

        if not regulatory_interactions:
            regulatory_interactions = {}

        if not targets:
            targets = {}

        if not regulators:
            regulators = {}

        self._id = None
        self._name = None
        self._expression_coef = None
        self._aliases = None
        self._regulatory_interactions = None
        self._targets = None
        self._regulators = None
        self._model = None
        self._is_regulator = None
        self._is_target = None
        self._solution = None

        self.__build_regulatory_variable__(identifier,
                                           name,
                                           expression_coef,
                                           aliases,
                                           regulatory_interactions,
                                           targets,
                                           regulators,
                                           model,
                                           is_regulator,
                                           is_target)

    def __str__(self):
        return self.name

    def __repr__(self):

        return self.id

    # def __repr__(self):
    #
    #     return '{}({}, {}, {}) at {}'.format(self.__class__.__name__, self._id, self._name, self._aliases, id(self))

    def __delete__(self, instance):
        del self.id
        del self.name
        del self.expression_coef
        del self.aliases
        del self.regulatory_interactions
        del self.targets
        del self.regulators
        del self.model
        del self.is_regulator
        del self.is_target
        del self.solution

    def __build_regulatory_variable__(self,
                                      identifier,
                                      name,
                                      expression_coef,
                                      aliases,
                                      regulatory_interactions,
                                      targets,
                                      regulators,
                                      model,
                                      is_regulator,
                                      is_target):

        if not identifier:
            raise TypeError("An ID must be provided")

        if not isinstance(regulatory_interactions, dict):
            raise TypeError(
                "regulatory interactions must be a dict of RegulatoryInteraction objects")

        from mewpy.regulation.regulatory_interaction import RegulatoryInteraction

        if regulatory_interactions and not isinstance(regulatory_interactions[list(regulatory_interactions.keys())[0]],
                                                      RegulatoryInteraction):
            raise TypeError(
                "regulatory interactions must be a dict of RegulatoryInteraction objects")

        if not isinstance(regulators, dict):
            raise TypeError(
                "Regulators must be a dict of RegulatoryVariables objects")

        if regulators and not isinstance(regulators[list(regulators.keys())[0]], RegulatoryVariable):
            raise TypeError(
                "Regulators must be a dict of RegulatoryVariables objects")

        if not isinstance(targets, dict):
            raise TypeError(
                "Targets must be a dict of RegulatoryVariables objects")

        if targets and not isinstance(targets[list(targets.keys())[0]], RegulatoryVariable):
            raise TypeError(
                "Targets must be a dict of RegulatoryVariables objects")

        self.id = identifier
        self.aliases = aliases
        self.name = name
        self.expression_coef = expression_coef

        self._regulatory_interactions = regulatory_interactions
        self._targets = targets
        self._regulators = regulators

        self.is_regulator = is_regulator
        self.is_target = is_target

        self.model = model

    @property
    def id(self):
        return getattr(self, '_id', None)

    @property
    def name(self):
        return getattr(self, '_name', None)

    @property
    def expression_coef(self):
        return getattr(self, '_expression_coef', None)

    @property
    def aliases(self):
        return getattr(self, '_aliases', None)

    @property
    def regulatory_interactions(self):
        return getattr(self, '_regulatory_interactions', None)

    @property
    def targets(self):
        return getattr(self, '_targets', None)

    @property
    def regulators(self):
        return getattr(self, '_regulators', None)

    @property
    def conditions(self):
        return getattr(self, '_conditions', None)

    @property
    def model(self):
        return getattr(self, '_model', None)

    @property
    def is_regulator(self):
        return getattr(self, '_is_regulator', None)

    @property
    def is_target(self):
        return getattr(self, '_is_target', None)

    @property
    def solution(self):
        return getattr(self, '_solution', None)

    @id.setter
    def id(self, value):

        if value == self._id:
            pass

        elif not isinstance(value, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        else:

            if self.model:

                if value in self.model.regulators and value in self.model.targets:
                    def reg_id(message):
                        warn(message, UserWarning, stacklevel=2)

                    reg_id("The ID already exists in the dictionary of regulators and targets of the regulatory "
                           "model. Ignoring the alteration")

                else:

                    if self._id in self.model.regulators:
                        del self.model.regulators[self._id]
                        self.model.regulators[value] = self

                    if self._id in self.model.targets:
                        del self.model.targets[self._id]
                        self.model.targets[value] = self

                    if self.regulatory_interactions:
                        for reg_inter in self.regulatory_interactions_gen():

                            if self._id in reg_inter.regulators:
                                del reg_inter.regulators[self._id]
                                reg_inter.regulators[value] = self

                    if self.regulators:
                        for reg in self.regulators.values():
                            del reg.targets[self._id]
                            reg.targets[value] = self

                    if self.targets:
                        for tar in self.targets.values():
                            del tar.regulators[self._id]
                            tar.regulators[value] = self

                    self._id = value

            else:

                if self.regulatory_interactions:

                    for reg_inter in self.regulatory_interactions_gen():

                        if self._id in reg_inter.regulators:
                            del reg_inter.regulators[self._id]
                            reg_inter.regulators[value] = self

                self._id = value

    @name.setter
    def name(self, value):

        if value == self._name:
            pass

        elif not isinstance(value, str):
            raise TypeError("The name must be a string. Use str() to convert")

        else:
            self._name = value

    @expression_coef.setter
    def expression_coef(self, value):

        if value == self._expression_coef:
            pass

        elif not isinstance(value, int) and not isinstance(value, float) and value is not None:
            raise TypeError("The expression coefficient must be a number (int or float) or None. "
                            "Use float() to convert")

        else:
            self._expression_coef = value

    @aliases.setter
    def aliases(self, value):

        if value == self._aliases:
            pass

        elif not value:
            pass

        else:

            if not isinstance(value, list):
                raise TypeError("The aliases must be a list. Use list() to convert")

            if value and not isinstance(value[0], str):
                raise TypeError("The aliases must be a list. Use list() to convert")

            if not self._aliases:
                self._aliases = []

            for val in value:
                if val not in self._aliases:
                    self._aliases = self._aliases + [val]

    @regulatory_interactions.setter
    def regulatory_interactions(self, value):

        if value == self._regulatory_interactions:
            pass

        else:
            raise BaseException("Regulatory interactions property cannot be changed. "
                                "To add new regulatory interactions to the model, "
                                "please use the add_new_regulatory_interaction method.")

    @targets.setter
    def targets(self, value):

        if value == self._targets:
            pass

        else:
            raise BaseException("Targets property cannot be changed. "
                                "To add new targets to the model, please use the add_new_target method.")

    @regulators.setter
    def regulators(self, value):

        if value == self._regulators:
            pass

        else:
            raise BaseException("Regulators property cannot be changed. "
                                "To add new regulators to the model, please use the add_new_regulator method.")

    @model.setter
    def model(self, value):

        # TODO: The setter should also remove the regulatory variable from the old model if exists

        from mewpy.regulation.regulatory_model import RegulatoryModel

        if value == self._model:
            pass

        elif not value:
            self._model = None

        elif not isinstance(value, RegulatoryModel):
            raise TypeError(
                "Regulatory model must be a Regmodel object or None")

        else:

            # TODO: This should point all variables and interactions related with this one to the same model, but
            #             #  this might be a recursion problem

            if not self.is_target and not self.is_regulator:
                def model_id(message):
                    warn(message, UserWarning, stacklevel=2)

                model_id("The regulatory variable must be regulator or target to be added to the regulatory model. "
                         "Ignoring the alteration")

            elif self.id not in value.regulators and self.is_regulator:
                value.regulators[self.id] = self
                self._model = value

            elif self.id not in value.targets and self.is_target:
                value.targets[self.id] = self
                self._model = value

            else:
                def model_id(message):
                    warn(message, UserWarning, stacklevel=2)

                model_id("The ID already exists in the dictionary of regulators and targets of the regulatory model. "
                         "Ignoring the alteration")

    @is_regulator.setter
    def is_regulator(self, value):

        if value == self._is_regulator:
            pass

        elif not isinstance(value, bool):
            raise TypeError(
                "is_regulator must be a boolean. Use bool() to convert")

        else:
            self._is_regulator = value

            if self.model is not None:

                if self.id not in self.model.regulators:
                    self.model.regulators[self.id] = self

    @is_target.setter
    def is_target(self, value):

        if value == self._is_target:
            pass

        elif not isinstance(value, bool):
            raise TypeError(
                "is_target must be a boolean. Use bool() to convert")

        else:
            self._is_target = value

            if self.model is not None:

                if self.id not in self.model.targets:
                    self.model.targets[self.id] = self

    @solution.setter
    def solution(self, value):

        if value == self._solution:
            pass

        else:
            raise BaseException("Solution property cannot be changed. "
                                "This property is fulfilled when the regulatory model is simulated.")

    @id.deleter
    def id(self):
        del self._id

    @name.deleter
    def name(self):
        del self._name

    @aliases.deleter
    def aliases(self):
        del self._aliases

    @regulatory_interactions.deleter
    def regulatory_interactions(self):
        del self._regulatory_interactions

    @expression_coef.deleter
    def expression_coef(self):
        del self._expression_coef

    @is_regulator.deleter
    def is_regulator(self):
        del self._is_regulator

    @is_target.deleter
    def is_target(self):
        del self._is_target

    @model.deleter
    def model(self):
        del self._model

    @regulators.deleter
    def regulators(self):
        del self._regulators

    @targets.deleter
    def targets(self):
        del self._targets

    @solution.deleter
    def solution(self):
        del self._solution

    def regulators_gen(self):

        return (value for value in self._regulators.values())

    def targets_gen(self):

        return (value for value in self._targets.values())

    def regulatory_interactions_gen(self):

        return (value for value in self._regulatory_interactions.values())

    def knock_out(self):

        self.expression_coef = 0

    def remove_from_model(self):

        if self.model is not None:

            self.model.remove_regulatory_variable(self.id)

        else:
            def remove_from_model_reg(message):
                warn(message, UserWarning, stacklevel=2)

            remove_from_model_reg("The regulatory variable is not bounded to any regulatory model")



from mewpy.regulation import RFBAModel
import time
import warnings
import os

warnings.filterwarnings("ignore")


def sample_network(dynamic=True):
    """
    Sample network available in examples using cobra as reader

    :return: sample sr-fba model
    """

    # Sample network

    import cobra.io
    from mewpy.simulation.cobra import Simulation
   
    DIR = os.path.dirname(os.path.realpath(__file__)) 
    cbm_model_f = os.path.join(DIR,'../../../examples/models/regulation/SampleNet.xml')
    reg_model_f = os.path.join(DIR,'../../../examples/models/regulation/SampleRegNet.csv')

    _BIOMASS_ID = 'r11'

    t0 = time.time()
    model = cobra.io.read_sbml_model(cbm_model_f)
    model.objective = _BIOMASS_ID

    simulation = Simulation(model)
    t1 = time.time()

    print(t1 - t0)

    t0 = time.time()
    rfba = RFBAModel.from_tabular_format(reg_model_f, model, simulation,
                                         sep=',', id_col=0, rule_col=1, header=None)
    t1 = time.time()

    print(t1 - t0)

    initial_state = {var: 1 for var in rfba.metabolic_regulatory_genes}

    # r15
    initial_state.update({
        'pH': 7,
        'g21': 1,
        'g22': 0,
        'g23': 1,
        'g24': 1,
        'g25': 1,
        'g26': 1,
        'g27': 1,
        'g28': 1,
        'g29': 1,
        'g30': 0,
        'g31': 0,
        'g32': 0,
        'g33': 1,
        'g36': 1
    })

    # # not r15
    # initial_state.update({
    #     'pH': 7,
    #     'g21': 1,
    #     'g22': 0,
    #     'g23': 1,
    #     'g24': 0,
    #     'g25': 0,
    #     'g26': 0,
    #     'g27': 1,
    #     'g28': 1,
    #     'g29': 1,
    #     'g30': 0,
    #     'g31': 1,
    #     'g32': 0,
    #     'g33': 1,
    #     'g36': 0,
    # })

    initial_state.update({
        'pH': 7
    })

    rfba.initial_state = initial_state

    if dynamic:
        t0 = time.time()
        rfba.dynamic_simulate()
        t1 = time.time()

    else:

        t0 = time.time()
        rfba.simulate()
        t1 = time.time()

    print(t1 - t0)

    return rfba


def cobra_ecoli_core_model(dynamic=True):
    """

    RFBA COBRA Ecoli Core Model (subset of the genome-scale metabolic reconstruction iAF1260)

    Available at Orth J, Fleming R, Palsson B. 2010. Reconstruction and Use of Microbial Metabolic Networks:
    the Core Escherichia coli Metabolic Model as an Educational Guide, EcoSal Plus 2010;
    doi:10.1128/ecosalplus.10.2.1

    Ecoli core model from cobra.test
    Regulatory model from http://systemsbiology.ucsd.edu/Downloads/EcoliCore

    :return: rfba model
    """

    import cobra.test
    from mewpy.simulation.cobra import Simulation

    DIR = os.path.dirname(os.path.realpath(__file__))
    reg_model_f = os.path.join(DIR,'../../../examples/models/regulation/core_TRN_v2.csv')
    aliases_f = os.path.join(DIR,'../../../examples/models/regulation/core_TRN_rfba_aliases.csv')

    _BIOMASS_ID = 'Biomass_Ecoli_core'
    _O2 = 'EX_o2_e'
    _GLC = 'EX_glc__D_e'
    _FUM = 'EX_fum_e'
    _AC = 'EX_ac_e'
    _GLU = 'EX_glu__L_e'
    _LAC = 'EX_lac__D_e'
    _SUC = 'EX_succ_e'

    t0 = time.time()
    model = cobra.test.create_test_model("textbook")
    model.objective = _BIOMASS_ID

    envcond = {_GLC: (-10.0, 100000.0),
               _O2: (-1000, 100000.0),
               _FUM: (0, 0.0),
               _AC: (0, 0.0),
               _GLU: (0, 0.0),
               _LAC: (0, 0.0),
               _SUC: (0, 0.0)}

    simulation = Simulation(model, envcond=envcond)
    t1 = time.time()

    print(t1 - t0)

    t0 = time.time()
    rfba = RFBAModel.from_tabular_format(reg_model_f, model, simulation,
                                         sep=',', id_col=1, rule_col=2, aliases_cols=[0], header=0)
    rfba.update_aliases_from_tabular_format_file(aliases_f, id_col=1, aliases_cols=[0])
    t1 = time.time()

    print(t1 - t0)

    if dynamic:
        t0 = time.time()
        rfba.dynamic_simulate()
        t1 = time.time()

    else:

        t0 = time.time()
        rfba.simulate()
        t1 = time.time()

    print(t1 - t0)

    return rfba


def framed_imc1010_model(dynamic=True):
    """

    RFBA FRAMED Ecoli iMC1010 model from Covert et al. 2004 at https://doi.org/10.1038/nature02456
    This model uses the E. coli metabolic model iJR904 available at https://www.ebi.ac.uk/biomodels/MODEL1507180060
    and published at https://doi.org/10.1186/gb-2003-4-9-r54

    Some rules had to be adjusted though iJR904 had to be adjusted, as it didn't match SR_FBA original publication or had errors

    The following reactions were added as in the original publication:
    
        - h2so: 2 o2_c + h2s_c -> (0, 999999) so4_c + 2 h_c
        - h2st: h2s_e <-> (-999999, 999999) h2s_c
        - h2s_ext: h2s_e -> (0, 999999)
        - 5dglcn_ext: 5dglcn_e -> (0, 999999)
        - btn_ext: btn_e -> (0, 999999)
        - cbi_ext: cbi_e -> (0, 999999)
        - h2o2_ext: h2o2_e -> (0, 999999)
        - ppa_ext: ppa_e -> (0, 999999)
        - thym_ext: thym_e-> (0, 999999)


    The following reactions were removed from iJR904, as they were duplicated or wrong:
        - GALU (duplicated GALUi (correct GPR))


    The following reactions were removed from SR-FBA orginal publication model, as they were duplicated or wrong:
        - ptrca (duplicated PTRCTA)
        - indolet (unbalanced and duplicated indolet2r)


    The following indirect mappings were found:

        cbiat, R_CBIAT_DELETE
        cblat, R_CBLAT_DELETE
        d-lact2, R_D_LACt2
        dhptdc, R_DHPTDCs
        l-lacd2, R_L_LACD2
        l-lact2r, R_L_LACt2r
        lcar, R_LCARS
        nh3t, R_NH4t
        l-lacd3, R_L_LACD3
        test_akgdh, R_AKGDH
        test_nadtrhd, R_NADTRHD
        trpas1, CYSDS

    Others:
        - iJR904 has an additional reaction to ptrcta, namely ORNTA


    Target genes with empty rules were set as ON in version 6. Alternatively, the state of all target genes can be set to 1 using the initial state setter

    :return: rfba model
    
    """

    from reframed.io.sbml import load_cbmodel
    from mewpy.simulation.reframed import Simulation

    DIR = os.path.dirname(os.path.realpath(__file__)) 
    cbm_model_f = os.path.join(DIR,"../../../examples/models/regulation/iJR904_srfba.xml")
    reg_model_f = os.path.join(DIR,'../../../examples/models/regulation/imc1010_v6.csv')
    aliases_f = os.path.join(DIR,'../../../examples/models/regulation/imc1010_rfba_aliases.csv')
    # env_cond_f = os.path.join(DIR,"../../../examples/models/regulation/imc1010_env_cond.xlsx")

    _BIOMASS_ID = 'R_BiomassEcoli'
    _O2 = 'R_EX_o2_e'
    _GLC = 'R_EX_glc_DASH_D_e'
    _CO2 = 'R_EX_co2_e'
    _FE2 = "R_EX_fe2_e"
    _H = "R_EX_h_e"
    _H2O = "R_EX_h2o_e"
    _K = "R_EX_k_e"
    _NA1 = "R_EX_na1_e"
    _NH4 = "R_EX_nh4_e"
    _PI = "R_EX_pi_e"
    _SO4 = "R_EX_so4_e"
    _SUCC = "R_EX_succ_e"

    _GLY = "R_EX_gly_e"
    _DALA = "R_EX_ala_DASH_D_e"

    t0 = time.time()
    model = load_cbmodel(cbm_model_f)
    model.set_objective({_BIOMASS_ID: 1})

    # Wild-type
    # import pandas as pd
    # envcond = pd.read_excel(env_cond_f)
    # envcond = {j: (float(envcond.iloc[i, 1]), float(envcond.iloc[i, 2])) for i, j in enumerate(envcond.iloc[:, 0])}

    # envcond = {_GLC: (-10.0, 100000.0),
    #            _O2: (-10.0, 100000.0),
    #            _CO2: (-15.0, 100000.0),
    #            _FE2: (-10.0, 100000.0),
    #            _H: (-10.0, 100000.0),
    #            _H2O: (-55.0, 100000.0),
    #            _K: (-10.0, 100000.0),
    #            _NA1: (-10.0, 100000.0),
    #            _NH4: (-10.0, 100000.0),
    #            _PI: (-15.0, 100000.0),
    #            _SO4: (-10.0, 100000.0),
    #            _SUCC: (0.0, 100000.0)}

    envcond = {_GLC: (-10.0, 100000.0),
               _SUCC: (0.0, 100000.0),
               _NH4: (-10.0, 100000.0),
               _O2: (-10.0, 100000.0),
               _CO2: (-15.0, 100000.0),
               _PI: (-15.0, 100000.0),
               _SO4: (-10.0, 100000.0),
               _H: (-10.0, 100000.0),
               _H2O: (-55.0, 100000.0)}

    # envcond.update({_GLC: (-10.0, 100000.0),
    #                 _SUCC: (0.0, 100000.0),
    #                 _NH4: (-10.0, 100000.0),
    #                 _O2: (-10.0, 100000.0),
    #                 _CO2: (-15.0, 100000.0),
    #                 _PI: (-15.0, 100000.0),
    #                 _SO4: (-10.0, 100000.0),
    #                 _H: (-10.0, 100000.0),
    #                 _H2O: (-55.0, 100000.0)})

    # Paper analysis encoding: in vivo/FBA/RFBA;  (+ growth) (- no growth)

    # OD600 growth on Glycine + b3359 M ARGDR
    # In paper: -/-/-

    # envcond.update(
    #     {_GLC: (0.0, 100000.0),
    #      _GLY: (-10.0, 100000.0),
    #      'R_ACOTA': (0.0,0.0)}
    # )

    # OD600 growth on Glycine + b3572 M AVTA1
    # In paper: -/+/-

    # envcond.update(
    #     {_GLC: (0.0, 100000.0),
    #      _GLY: (-10.0, 100000.0),
    #      'R_VPAMT': (0.0,0.0)}
    # )

    # OD600 growth on Glycine + b3671 M ILVB1 b3671 M ILVB2
    # In paper: +/+/-

    # envcond.update(
    #     {_GLC: (0.0, 100000.0),
    #      _GLY: (-10.0, 100000.0)}
    # )

    # OD600 growth on D-Alanine + b3423 R GlpR
    # In paper: +/not observable/+
    # SET GlpR to false in csv

    # envcond.update(
    #     {_GLC: (0.0, 100000.0),
    #      _DALA: (-10.0, 100000.0)}
    # )

    # OD600 growth on D-Alanine + b3119 R TdcR
    # In paper: +/not observable/+
    # SET TdcR to false in csv

    # envcond.update(
    #     {_GLC: (0.0, 100000.0),
    #      _DALA: (-10.0, 100000.0)}
    # )

    # OD600 growth on D-Alanine + b3572 M AVTA1
    # In paper: +/not observable/+

    # envcond.update(
    #     {_GLC: (0.0, 100000.0),
    #      _DALA: (-10.0, 100000.0),
    #      'R_VPAMT': (0.0,0.0)}
    # )

    simulation = Simulation(model, envcond=envcond)
    t1 = time.time()

    print(t1 - t0)

    t0 = time.time()
    rfba = RFBAModel.from_tabular_format(reg_model_f, model, simulation,
                                         sep=';', id_col=1, rule_col=4, aliases_cols=[0, 2, 3], header=0)
    rfba.update_aliases_from_tabular_format_file(aliases_f, sep=';', id_col=0, aliases_cols=[1])
    t1 = time.time()

    print(t1 - t0)

    initial_state = {var: 1 for var in rfba.targets}
    initial_state.update({_BIOMASS_ID: 0.003})
    rfba.initial_state = initial_state

    if dynamic:
        t0 = time.time()
        rfba.dynamic_simulate()
        t1 = time.time()

    else:

        t0 = time.time()
        rfba.simulate()
        t1 = time.time()

    print(t1 - t0)

    return rfba


if __name__ == '__main__':
    rfba_sample_net = sample_network()
    rfba_ecoli_core = cobra_ecoli_core_model()
    rfba_ecoli_imc1010 = framed_imc1010_model()





class ModelConstants:
    # Default reactions upper bound (used instead of Inf)
    REACTION_UPPER_BOUND = 10000
    # reset solver
    RESET_SOLVER = False


class EAConstants:
    # Default fitness value for invalid solution on minimization problems
    MAX_THRESHOLD = 100000000.0
    # Default fitness value for invalid solution on maximization problems
    MIN_THRESHOLD = 0.0
    # Default solution minimum size (minimum number of reactions/proteins/genes to KO/OU)
    MIN_SOLUTION_SIZE = 1
    # Default solution maximum size (maximum number of reactions/proteins/genes to KO/OU)
    MAX_SOLUTION_SIZE = 10
    # Default OU levels (Note: 0 levels are interpreted as KO)
    LEVELS = [0, 1/32, 1/16, 1/8, 1/4, 1/2, 2, 4, 8, 16, 32]
    # Default OU levels for OptRAM(Note: instead of 0 use 0.001 as a logaritmic function is used to process expression values)
    OPTRAM_LEVELS = [0.001, 1/32, 1/16, 1/8, 1/4, 1/2, 2, 4, 8, 16, 32]
    # Number of cpus for multiprocessor evaluation of candidates.
    # If NUM_CPUS is non positive, half of the available cpus are used
    NUM_CPUS = -1
    # Maximum number of generations (used as stopping criteria for the EA)
    MAX_GENERATIONS = 100
    # DEBUG
    DEBUG = False



"""
##################################################################

GECKO over REFRAMED

##################################################################
"""
from mewpy.model.gecko import GeckoModel, ModelList
from mewpy.simulation.reframed import GeckoSimulation
from mewpy.simulation import SimulationMethod
from collections import OrderedDict
import statistics
import pandas as pd
import os

def simulation_one():
    """
    Identify all pairs of (reaction,reaction_REV) associated to each protein draw reaction
    """
    model = GeckoModel('single-pool')
    simulation = GeckoSimulation(model)
    
    result= simulation.simulate(method= SimulationMethod.pFBA)
    wt_fluxes = result.fluxes

    rev_pairs = model.protein_rev_reactions
    with open("protein_reaction_under.csv",'w') as f:
        f.write("rxn; rxn_REV ; ; WT_rxn_flux; WT_rxn_REV_flux; ; O_rxn_flux;O_rxn_REV_flux\n")
        for protein_id in rev_pairs.keys():
            constraints = OrderedDict()
            rxn ='draw_prot_{}'.format(protein_id)
            constraints[rxn]= (0,0.5*wt_fluxes[rxn]) 
            result= simulation.simulate(constraints= constraints, method= SimulationMethod.pFBA)
            ssfluxes = result.fluxes
            if ssfluxes:
                f.write(protein_id+";;;;;;\n")
                for r_id, r_rev_id in rev_pairs[protein_id]:    
                    f.write("{}; {}; ; {}; {}; ;{}; {}\n".format(r_id,r_rev_id,wt_fluxes[r_id],wt_fluxes[r_rev_id],ssfluxes[r_id],ssfluxes[r_rev_id]))


def simulation_two():
    """
    Saves essential proteins to file
    """
    model = GeckoModel('single-pool')
    simulation = GeckoSimulation(model)
    essential_prot = simulation.essential_proteins('draw_prot_')
    with open('target-prot-single-pool.txt','w') as f:
        for p in model.proteins:
            if p not in essential_prot:
                f.write(p+"\n")




def simulation_three():
    constraints ={'draw_prot_Q04396': (2.3029771980628338e-07,1000),'draw_prot_P27515':(0,0.0),'draw_prot_P53687':(0.0,1000),'draw_prot_P81450': 0,
                  'draw_prot_P41939': (3.944331158229099e-06,1000),	'r_0659_REVNo1':0,'draw_prot_P19657':(0,0.0),'draw_prot_P37254':(0.0,1000),
                  'draw_prot_P33317': (0,0.0),'draw_prot_P54114':(0,0.0),'draw_prot_P25373':(0.0,1000),	'draw_prot_P08067':(0,2.1621430284827036e-06),
                  'draw_prot_P39002': (0.0,1000),'draw_prot_Q04792':(0.0,1000),	'draw_prot_Q12320':	(0.0,1000),	'draw_prot_P07347':(0.0,1000), 'draw_prot_P00958':(0,3.4862363761891556e-06),
                  'draw_prot_P15179': (0,0.0), 'draw_prot_P53332':(0.0,1000),'draw_prot_P23202':(0.0,1000),	'draw_prot_P21672':	(0,0.0),'draw_prot_P32191':(0,0.0),
                  'draw_prot_P38071': 0,'draw_prot_P06169':	(0.00019588347980199452,1000),'draw_prot_P22803':0,	'draw_prot_P38986':(0.0,1000),
                  'draw_prot_P32621': (0.0,1000),'draw_prot_P27472':(1.7311263946838522e-07,1000),'draw_prot_P00549': (0,1.624598159388334e-05),
                  'draw_prot_P32796':0}
    model = GeckoModel('single-pool')
    model.set_objective({'r_2111': 0.0 , 'r_4041':1.0})
    simulation = GeckoSimulation(model)
    result = simulation.simulate(method=SimulationMethod.pFBA)
    reference = result.fluxes
    
    result = simulation.simulate(constraints=constraints)
    
    from mewpy.optimization.evaluation import WYIELD, BPCY, TargetFlux
    evaluator_1 = WYIELD("r_2111", "r_2056")
    print(evaluator_1.get_fitness(result,None))
    evaluator_2 = BPCY("r_2111", "r_2056", "r_1714_REV", method=SimulationMethod.lMOMA ,reference=reference)
    print(evaluator_2.get_fitness(result,None))
    evaluator_3 = TargetFlux("r_2056")
    print(evaluator_3.get_fitness(result,None))


def test_basic_gecko_adjustment():
    in_model = {'P00549': 0.1, 'P31373': 0.1, 'P31382': 0.1, 'P39708': 0.1, 'P39714': 0.1, 'P39726': 0.1, 'Q01574': 0.1}
    not_in_model = {'P10591': 0.1, 'P31383': 0.1, 'P32471': 0.1}
    measurements = pd.concat([pd.Series(in_model), pd.Series(not_in_model)])
    model = GeckoModel('multi-pool')
    model.limit_proteins(fractions=pd.Series(measurements))
    simul = GeckoSimulation(model)
    sol = simul.simulate()
    assert sol.objective_value > 0.05
    assert len(model.proteins) - len(model.pool_proteins) - len(in_model) == 0
    assert all(model.reactions[rxn].ub > 0 for rxn in model.individual_protein_exchanges)




def test_gecko_adjustment_sanchez_etal():
    mmol_gdw = pd.read_csv(os.path.join(os.path.dirname(__file__), '../model/data/sanchez-mmol_gdw.csv'))
    PROTEIN_PROPERTIES = ModelList().protein_properties()
    ggdw = pd.Series(PROTEIN_PROPERTIES.loc[mmol_gdw.index, 'mw'] / 1000.) * pd.Series(mmol_gdw)
    model = GeckoModel('multi-pool')
    simulation = GeckoSimulation(model)
    result = simulation.simulate()
    growth_rate_unlimited_protein = result.objective_value
    model.limit_proteins(ggdw=pd.Series(ggdw))
    result = simulation.simulate()
    growth_rate_limited_protein = result.objective_value
    # should be smaller, but how much..
    assert growth_rate_limited_protein < 0.8 * growth_rate_unlimited_protein
    measured_in_model = set(mmol_gdw.index).intersection(model.proteins)
    assert sum(model.concentrations[p] - ggdw[p] for p in measured_in_model) < 1e-10
    #assert sum(abs(rxn.upper_bound - mmol_gdw[rxn.metadata['uniprot']])
    #           for rxn in model.individual_protein_exchanges) < 1e-6
    #assert sum(rxn.metabolites[model.common_protein_pool] +
    #           PROTEIN_PROPERTIES.loc[rxn.annotation['uniprot'], 'mw'] / 1000.
    #           for rxn in model.pool_protein_exchanges) < 1e-6
    #assert model.p_measured > 0.25                                  # With yeast 8.1.3 -> p_measured = 0.296
    #assert model.f_mass_fraction_measured_matched_to_total > 0.25   # With yeast 8.1.3 -> f = 0.304
    #assert model.protein_pool_exchange.upper_bound > 0.015          # With yeast 8.1.3 -> pool_exchange = 0.0212



def simulation_four():

    constraints = {'draw_prot_P17505':0,
                   'draw_prot_P54885':0,
                   'draw_prot_P50107':0,
                   'draw_prot_P37303':0,		
                   'draw_prot_P38113':0,
                   'draw_prot_Q06408':0,
                   'draw_prot_P32340':0,
                   'draw_prot_P00817':0,
                   'draw_prot_P09440':0,
                   'draw_prot_P42951':0,
                   'draw_prot_P36013':0,
                   'draw_prot_P32473':0,
                   'draw_prot_P27680':0,		
                   'draw_prot_P41939':0,
                   'draw_prot_P17695':0,
                   'draw_prot_P32383':0,		
                   'draw_prot_P38840':0,
                   'draw_prot_P38715':0,
                   'draw_prot_P23542':0,		
                   'draw_prot_Q01574':0,		
                   'draw_prot_P06208':0,	
                   'draw_prot_P00330':0,	
                   'draw_prot_P33330':0,		
                   'draw_prot_P32419':0,		
                   'draw_prot_P47143':0,		
                   'draw_prot_P32179':0,	
                   'draw_prot_P06169':0}
    
    model = GeckoModel('single-pool')
    model.set_objective({'r_2111': 0 , 'r_4041':1.0})
    simulation = GeckoSimulation(model)
    result = simulation.simulate(method = SimulationMethod.pFBA)
    reference = result.fluxes
    print("biomass {}   tyrosine: {}".format(reference['r_4041'],reference['r_1913']))
    result = simulation.simulate(method = SimulationMethod.pFBA,constraints=constraints)
    reference = result.fluxes
    print("biomass {}   tyrosine: {}".format(reference['r_4041'],reference['r_1913']))

    reactions = [] 
    for prot_exchange in constraints.keys():
        prot = prot_exchange[len('draw_prot_'):]
        reactions.append(simulation.protein_reactions(prot))
    reactions.sort()    
    print(reactions)
    print(len(reactions))


def gecko_ec():
    import os
    dir_path = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(dir_path, '../../../examples/models/gecko/')
    DATA_FILE = os.path.join(PATH, 'eciML1515_batch.xml')
    from reframed.io.sbml import load_cbmodel
    m = load_cbmodel(DATA_FILE)
    model = GeckoModel(m, biomass_reaction_id='R_BIOMASS_Ec_iML1515_core_75p37M',protein_pool_exchange_id='R_prot_pool_exchange',reaction_prefix='R_')
    model.set_objective({'R_BIOMASS_Ec_iML1515_core_75p37M': 1.0})
    
    # change protein pool bound (suggested by Leslie)
    # model.reactions['R_prot_pool_exchange'].ub = 0.26 
    from mewpy.simulation import get_simulator, SimulationMethod
 
    c = {'P32131': 0.125, 'P45425': 32, 'P0A6E1': 32, 'P0A9I8': 0, 'P52643': 4, 'P37661': 0.125}
    
    from mewpy.optimization.evaluation import BPCY, WYIELD
    from mewpy.problems import GeckoOUProblem
    # the evaluation (objective) functions
    evaluator_1 = BPCY("R_BIOMASS_Ec_iML1515_core_75p37M", 'R_EX_tyr__L_e',
                       method=SimulationMethod.pFBA)
    # FVA MAX is strangely very high... changing the default alpha (0.3) to compensate..
    evaluator_2 = WYIELD("R_BIOMASS_Ec_iML1515_core_75p37M", 'R_EX_tyr__L_e', alpha= 0.01 )
    
    # The optimization problem
    problem = GeckoOUProblem(model,
                              fevaluation=[evaluator_1, evaluator_2],
                              envcond={},
                              prot_prefix='R_draw_prot_',
                              candidate_max_size=30)
    

    #c1 = problem.translate(c,reverse=True)
    #print(c1)
    #c2 = problem.decode(c1)
    #print(c2)
    sim = get_simulator(model)
    
    #c2 = {'R_draw_prot_P0AFV4': 0.0, 'R_draw_prot_P67910': (0.0, 0.0), 'R_draw_prot_P02924': (0.0, 0.0), 'R_draw_prot_P07639': (1.1271272290940493e-07, 10000), 'R_draw_prot_P39172': 0.0, 'R_draw_prot_P0AER3': (0.0, 10000), 'R_draw_prot_P0A991': (0.0, 10000), 'R_draw_prot_P0ACD8': (0.0, 0.0), 'R_draw_prot_P00805': (0.0, 10000), 'R_draw_prot_P28635': (0.0, 0.0), 'R_draw_prot_P33593': (0.0, 0.0), 'R_draw_prot_P0A9H5': (0.0, 10000), 'R_draw_prot_P0A6L4': (0.0, 10000), 'R_draw_prot_P60560': (0.0, 10000), 'R_draw_prot_P37001': 0.0, 'R_draw_prot_P37355': (0.0, 0.0), 'R_draw_prot_P0AEE5': (0.0, 0.0), 'R_draw_prot_P0ABA0': (0.0, 0.0), 'R_draw_prot_P0ABK5': (0.0, 10000)}
    c2 = {'R_draw_prot_P69922': (0.0, 10000), 'R_draw_prot_P32176': (0.0, 0.0), 'R_draw_prot_P11349': (0.0, 10000), 'R_draw_prot_P37646': 0.0, 'R_draw_prot_P76577': (0.0, 0.0), 'R_draw_prot_P77788': (0.0, 0.0), 'R_draw_prot_P0AG20': (0.0, 10000), 'R_draw_prot_P63224': 0.0, 'R_draw_prot_P62623': (0.0, 3.28753677758505e-07), 'R_draw_prot_P10907': (0.0, 0.0), 'R_draw_prot_P0AER5': (0.0, 0.0), 'R_draw_prot_P27254': (0.0, 10000), 'R_draw_prot_P0A6E1': (4.0254906190734055e-05, 10000), 'R_draw_prot_P0A924': 0.0, 'R_draw_prot_P32055': (0.0, 10000), 'R_draw_prot_P21179': 0.0, 'R_draw_prot_P10378': 0.0}

    print("\nFBA")
    res = sim.simulate(method=SimulationMethod.FBA,constraints=c2)
    print(res)
    print("Biomass:",res.fluxes['R_BIOMASS_Ec_iML1515_core_75p37M'])
    print("TYR:",res.fluxes['R_EX_tyr__L_e'])
    
    print("\npFBA")
    res = sim.simulate(method=SimulationMethod.pFBA,constraints=c2)
    print(res)
    print("Biomass:",res.fluxes['R_BIOMASS_Ec_iML1515_core_75p37M'])
    print("TYR:",res.fluxes['R_EX_tyr__L_e'])
    
    print("\nlMOMA")
    res = sim.simulate(method=SimulationMethod.lMOMA,constraints=c2)
    print(res)
    print("Biomass:",res.fluxes['R_BIOMASS_Ec_iML1515_core_75p37M'])
    print("TYR:",res.fluxes['R_EX_tyr__L_e'])

    print("\nMOMA")
    res = sim.simulate(method=SimulationMethod.MOMA,constraints=c2)
    print(res)
    print("Biomass:",res.fluxes['R_BIOMASS_Ec_iML1515_core_75p37M'])
    print("TYR:",res.fluxes['R_EX_tyr__L_e'])

    print("\nFVA")
    res = sim.FVA(reactions=['R_EX_tyr__L_e'],constraints=c2)
    print(res)


if __name__ == "__main__":
    simulation_one()
    #simulation_two()
    #simulation_three()
    #test_basic_gecko_adjustment()
    #test_gecko_adjustment_sanchez_etal()
    #simulation_four()
    
    #gecko_ec()
    



from abc import abstractmethod
from operator import add, sub, mul, truediv, pow
from sympy.parsing.sympy_parser import parse_expr
from sympy import Symbol
import sys
import re


# Boolean operator symbols
S_AND = '&'
S_OR = '|'
S_NOT = '~'
S_ON = '1'
S_OFF = '0'
S_GREATER = '>'
S_LESS = '<'
S_EQUAL = '='
S_GREATER_THAN_EQUAL = '>='
S_LESS_THAN_EQUAL = '=>'

# Empty leaf symbol
EMPTY_LEAF = '@'

BOOLEAN_OPERATORS = [S_AND, S_OR, S_NOT]

BOOLEAN_STATES = [S_ON, S_OFF]

BOOLEAN_RELATIONALS = [S_GREATER, S_LESS, S_EQUAL]
BOOLEAN_EQUAL_RELATIONALS = [S_GREATER_THAN_EQUAL, S_LESS_THAN_EQUAL]

# TODO: This might be problematic, but it seems to be the better way. A disclaimer/instructions of how genes and
#  special vars must be formatted can also work. If the user provides a list of all genes/regulatory variables in the
#  TRN or search them in the ids column, this can also be some work arounds.

# Special chars to be replaced
BOOLEAN_SPECIAL_CHARS = {'-': '_dash_',
                         # '.' : '_dot_',
                         ',': '_comma_',
                         ';': '_semicolon_',
                         '[': '_lbracket_',
                         ']': '_rbracket_',
                         ':': '_colon_',
                         '+': '_plus_',
                         '/': '_slash_',
                         '\\"': '_backslash_',
                         '=': '_equal_',
                         '$': '_dollar_',
                         '%': '_percentage_',
                         '"': '_quotes_',
                         '(e)': '_e_',
                         '(c)': '_c_',
                         '(p)': '_p_',
                         '(E)': '_E_',
                         '(C)': '_C_',
                         '(P)': '_P_',
                         'lambda': '_lambda_function_',
                         'Add': '_Add_',
                         'add': '_add_',
                         'Mul': '_Mul_',
                         'mul': '_mul_',
                         'Pow': '_Pow_',
                         'pow': '_pow_',
                         }

# Increases the system recursion limit for long expressions
sys.setrecursionlimit(100000)


def evaluate_expression(expression, variables):
    """ Evaluates a logical expression (containing variables, 'and','or','(' and ')')
    against the presence (True) or absence (False) of propositions within a list.
    The evaluation is achieved usind python native eval function.

    :param str expression: The expression to be evaluated.
    :param list variables: List of variables to be evaluated as True.
    :returns: A boolean evaluation of the expression.

    """
    expression = expression.replace('(', '( ').replace(')', ' )')
    # Symbol conversion not mandatory
    expression = expression.replace('and', '&').replace('or', '|')
    tokens = expression.split()
    sentence = []
    for token in tokens:
        if token in "&|()":
            sentence.append(token)
        elif token in variables:
            sentence.append("True")
        else:
            sentence.append("False")
    proposition = ' '.join(sentence)
    return eval(proposition)


def evaluate_expression_tree(expression, variables):
    """ Evaluates a logical expression (containing variables, 'and','or', 'not','(' , ')') against the presence (True)
    or absence (False) of propositions within a list.
    Assumes the correctness of the expression. The evaluation is achieved by means of a parsing tree.

    :param str expression: The expression to be evaluated.
    :param list variables: List of variables to be evaluated as True.
    :returns: A boolean evaluation of the expression.

    """
    t = build_tree(expression, Boolean)
    evaluator = BooleanEvaluator(variables)
    res = t.evaluate(evaluator.f_operand, evaluator.f_operator)
    return res


class Node(object):
    """
    Binary syntax tree node.

    :param value: The node value.
    :param left: The left node or None.
    :param right: The right node or None.
    """

    def __init__(self, value, left=None, right=None):
        if left is not None and not isinstance(left, Node):
            raise ValueError("Invalid left element")
        if right is not None and not isinstance(right, Node):
            raise ValueError("Invalid right element")
        self.value = value
        self.left = left
        self.right = right

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self.is_leaf():
            return str(self.value)
        else:
            return f"{str(self.value)} ( {str(self.left)} , {str(self.right)} )"

    def is_leaf(self):
        """
        :returns: True if the node is a leaf False otherwise. Both left and right are None.
        """
        return not self.left and not self.right

    def is_empty_leaf(self):
        """
        :returns: True if the node is an empty leaf False otherwise.
        """
        return self.value == EMPTY_LEAF

    def is_unary(self):
        """
        :returns: True if the node is a unary operation False otherwise.
        """
        return (self.left.is_empty_leaf() and not self.right.is_empty_leaf()) or (
                not self.left.is_empty_leaf() and self.right.is_empty_leaf())

    def is_binary(self):
        """
        :returns: True if the node is a binary operation False otherwise.
        """
        return not self.left.is_empty_leaf() and not self.right.is_empty_leaf()

    def get_operands(self):
        if self.is_leaf():
            if self.value == EMPTY_LEAF:
                return set()
            else:
                return {self.value}
        else:
            return self.left.get_operands().union(self.right.get_operands())

    def print_node(self, level=0):
        tabs = ""
        for _ in range(level):
            tabs += "\t"
        if self.is_leaf():
            if self.value != EMPTY_LEAF:
                print(tabs, f"|____ \033[1;31m{self.value}\033[0;0m")
            else:
                pass
        else:
            print(tabs, f"|____[\033[;1m\033[94m{self.value}\033[0;0m]")
        if self.left is not None:
            self.left.print_node(level + 1)
        if self.right is not None:
            self.right.print_node(level + 1)

    def evaluate(self, f_operand=None, f_operator=None):
        """
        Evaluates the expression using the f_operand and f_operator mapping functions
        """
        if f_operand is None or f_operator is None:
            return eval(str(self))
        elif self.is_leaf():
            return f_operand(self.value)
        else:
            return f_operator(self.value)(self.left.evaluate(f_operand, f_operator),
                                          self.right.evaluate(f_operand, f_operator))

    def get_conditions(self):
        """
        Retrieves the propositional conditions

        return: The list of conditions
        """
        ops = self.get_operands()
        return set([i for i in ops if is_condition(i)])


class Syntax:
    """Defines an interface for the tree syntax parsing
       with operators, their precedence and associativity.
    """

    @abstractmethod
    def is_operator(self, op):
        raise NotImplementedError

    @abstractmethod
    def is_greater_precedence(self, op1, op2):
        raise NotImplementedError

    @abstractmethod
    def associativity(self, op):
        raise NotImplementedError

    @staticmethod
    def arity(op):
        return 2

    @staticmethod
    def replace():
        return {}


class Arithmetic(Syntax):
    """Defines a basic arithmetic sintax.
    """
    @staticmethod
    def is_operator(op):
        return op in ['+', '-', '*', '/', '^']

    @staticmethod
    def is_greater_precedence(op1, op2):
        pre = {'+': 0, '-': 0, '*': 1, '/': 1, '^': 2}
        return pre[op1] >= pre[op2]

    @staticmethod
    def associativity(op):
        ass = {'+': 0, '-': 0, '*': 0, '/': 0, '^': 1}
        return ass[op]

    @staticmethod
    def arity(op):
        ar = {'+': 2, '-': 2, '*': 2, '/': 2, '^': 2}
        return ar[op]


class ArithmeticEvaluator:

    @staticmethod
    def f_operator(op):
        operators = {'+': add, '-': sub, '*': mul, '/': truediv, '^': pow}
        if op in operators.keys():
            return operators[op]
        else:
            raise ValueError(f"Operator {op} not defined")

    @staticmethod
    def f_operand(op):
        if is_number(op):
            return float(op)
        else:
            raise ValueError(f"Operator {op} not defined")


class Boolean(Syntax):
    """
    A boolean syntax parser where (NOT) is considered to
    be defined as a binary operator (NOT a) = (EMPTY NOT a)
    """

    @staticmethod
    def is_operator(op):
        return op in [S_AND, S_OR, S_NOT]

    @staticmethod
    def is_greater_precedence(op1, op2):
        pre = {S_AND: 0, S_OR: 0, S_NOT: 1}
        return pre[op1] >= pre[op2]

    @staticmethod
    def associativity(op):
        ass = {S_AND: 0, S_OR: 0, S_NOT: 1}
        return ass[op]

    @staticmethod
    def arity(op):
        ar = {S_AND: 2, S_OR: 2, S_NOT: 1}
        return ar[op]

    @staticmethod
    def replace():
        r = {'not': [EMPTY_LEAF, S_NOT], 'and': [S_AND],
             'or': [S_OR], '&': [S_AND], '|': [S_OR], '~': [EMPTY_LEAF, S_NOT]}
        return r

    SYMPY_REPLACE = {'not': S_NOT,
                     '!': S_NOT,
                     'and': S_AND,
                     'or': S_OR,
                     'on': S_ON,
                     'off': S_OFF,
                     'true': S_ON,
                     'false': S_OFF}

    _SYMPY_RELATIONAL_REPLACE = {'greater than or equal': S_GREATER_THAN_EQUAL,
                                 'less than or equal': S_LESS_THAN_EQUAL,
                                 'greater or equal': S_GREATER_THAN_EQUAL,
                                 'less or equal': S_LESS_THAN_EQUAL,
                                 'greater than': S_GREATER,
                                 'less than': S_LESS,
                                 'greater': S_GREATER,
                                 'less': S_LESS,
                                 '==': S_EQUAL,
                                 'equal': S_EQUAL}


class BooleanEvaluator:
    """A boolean evaluator.

    :param list true_list: Operands evaluated as True. Operands not in the list are evaluated as False
    :param dict variables: A dictionary mapping symbols to values. Used to evaluate conditions.

    """

    def __init__(self, true_list=[], variables={}):
        self.true_list = true_list
        self.vars = variables

    def f_operator(self, op):
        operators = {S_AND: lambda x, y: x and y, S_OR: lambda x, y: x or y, S_NOT: lambda x, y: not y}
        if op in operators.keys():
            return operators[op]
        else:
            raise ValueError(f"Operator {op} not defined")

    def f_operand(self, op):
        if op.upper() == 'TRUE' or op == '1' or op in self.true_list:
            return True
        elif is_condition(op):
            return eval(op, None, self.vars)
        else:
            return False

    def set_true_list(self, true_list):
        self.true_list = true_list


class GeneEvaluator:
    """An evaluator for genes expression.

    :param genes_value: A dictionary mapping genes to values. Gene not in the dictionary have a value of 1.
    :param and_operator: function to be applied instead of (and', '&') (not case sensitive)
    :param or_operator: function to be applied instead of ('or','|') (not case sensitive)
    """

    def __init__(self, genes_value, and_operator, or_operator, prefix=""):

        self.genes_value = genes_value
        self.and_operator = and_operator
        self.or_operator = or_operator
        self.prefix = prefix

    def f_operator(self, op):
        operators = {S_AND: self.and_operator, S_OR: self.or_operator}
        if op in operators.keys():
            return operators[op]
        else:
            raise ValueError(f"Operator {op} not defined")

    def f_operand(self, op):
        if op[len(self.prefix):] in self.genes_value:
            return self.genes_value[op]
        else:
            return 1


# Tree
def build_tree(exp, rules):
    """ Builds a parsing syntax tree for basic mathematical expressions

    :param exp: the expression to be parsed
    :param rules: Sintax definition rules
    """
    replace_dic = rules.replace()
    exp_ = tokenize_infix_expression(exp)
    exp_list = []
    for i in exp_:
        if i.lower() in replace_dic:
            exp_list.extend(replace_dic[i.lower()])
        else:
            exp_list.append(i)
    stack = []
    tree_stack = []
    predecessor = None
    for i in exp_list:
        if not (rules.is_operator(i) or i in ['(', ')']):
            if predecessor and not (rules.is_operator(predecessor) or predecessor in ['(', ')']):
                s = tree_stack[-1].value
                tree_stack[-1].value = s + " " + i
            else:
                t = Node(i)
                tree_stack.append(t)
        elif rules.is_operator(i):
            if not stack or stack[-1] == '(':
                stack.append(i)

            elif rules.is_greater_precedence(i, stack[-1]) and rules.associativity(i) == 1:
                stack.append(i)

            else:
                while stack and stack[-1] != '(' and rules.is_greater_precedence(stack[-1], i) and rules.associativity(
                        i) == 0:
                    popped_item = stack.pop()
                    t = Node(popped_item)
                    t1 = tree_stack.pop()
                    t2 = tree_stack.pop()
                    t.right = t1
                    t.left = t2
                    tree_stack.append(t)
                stack.append(i)

        elif i == '(':
            stack.append('(')

        elif i == ')':
            while stack[-1] != '(':
                popped_item = stack.pop()
                t = Node(popped_item)
                t1 = tree_stack.pop()
                t2 = tree_stack.pop()
                t.right = t1
                t.left = t2
                tree_stack.append(t)
            stack.pop()
        predecessor = i

    while stack:
        popped_item = stack.pop()
        t = Node(popped_item)
        t1 = tree_stack.pop()
        t2 = tree_stack.pop()
        t.right = t1
        t.left = t2
        tree_stack.append(t)

    t = tree_stack.pop()

    return t


def tokenize_infix_expression(exp):
    return list(filter(lambda x: x != '', exp.replace('(', ' ( ').replace(')', ' ) ').split(' ')))


def is_number(token):
    """ Returns True if the token is a number
    """
    return token.replace('.', '', 1).replace('-', '', 1).isnumeric()


def is_condition(token):
    """ Returns True if the token is a condition
    """
    regexp = re.compile(r'>|<|=')
    return bool(regexp.search(token))


# generic regulatory rule parser
def generic_regulatory_rule_parser(rule):
    exp = tokenize_infix_expression(rule)
    parsed_str = ''.join(exp)

    return parsed_str, exp


# special chars filter
def special_chars_filter(rule, special_chars=None):
    """
    For a given rule it parses out the special chars by replacing them with the corresponding values in the
    special_chars argument. By default, the global dictionary BOOL_SPECIAL_CHARS is used. This global dictionary can be
    altered for adding or removing special chars.


    :param rule: str, the regulatory rule
    :param special_chars: dict or None, special (or not) chars to be replaced. BOOL_SPECIAL_CHARS by default special_chars argument must have unique keys and unique values!!
    :returns: str, the new parsed rule
    :returns: dict, the occurred replacements {replace: special_char}
    """

    if special_chars is None:
        special_chars = BOOLEAN_SPECIAL_CHARS

    replaces = {}

    for special_char, replace in special_chars.items():

        if special_char in rule:
            replaces[replace] = special_char
            rule = rule.replace(special_char, replace)

    return rule, replaces


# initial digit filter
def starts_with_digit_filter(rule):
    """
    For a given regulatory rule it checks if all regulatory variables start with a digit (str.is_digit()) and parses
    out the regulatory variable by adding the prefix '_st_dg_'

    :param rule: str, the regulatory rule
    :return: tuple, the new parsed regulatory variable as a string object and the occurred replacements as dict object
    """

    replaces = {}
    rule = '(' + rule + ')'
    regexp = re.compile(r'[^a-zA-Z|^_][0-9]+[a-zA-Z]|[^a-zA-Z|^_][0-9]+_')
    res = list(regexp.finditer(rule))

    new_rule = ''

    last_nd = 0
    for i in range(len(res)):
        st = res[i].start() + 1
        nd = res[i].end()
        new_rule = new_rule + rule[last_nd:st] + '_st_dg_' + rule[st:nd]
        last_nd = res[i].end()
        replaces['_st_dg_'] = ''

    new_rule = new_rule + rule[last_nd:]

    return new_rule[1:-1], replaces


# relational filter
def single_relational_filter(expression):
    """
    For a given regulatory expression checks if it is a relational expression and identifies the regulatory
    variable. Otherwise, returns false

    :param expression: str, the regulatory expression
    :return: tuple, the regulatory variable as a string object and the condition enclosed by parenthesis, or False
    """

    replace_dict = Boolean._SYMPY_RELATIONAL_REPLACE
    for condition in replace_dict:
        if condition in expression:
            expression = expression.replace(condition, replace_dict[condition])
        elif condition.upper() in expression:
            expression = expression.replace(condition.upper(), replace_dict[condition])
        elif condition.title() in expression:
            expression = expression.replace(condition.title(), replace_dict[condition])
        else:
            continue

    # first let's handle GreaterEqual LessEqual
    regex_str = '|'.join([regex for regex in BOOLEAN_EQUAL_RELATIONALS])
    regexp = re.compile(regex_str)
    res = list(regexp.finditer(expression))

    if len(res) == 1:
        return expression[0:res[0].start()], '(' + expression + ')'
    elif len(res) == 2:
        return expression[res[0].end():res[1].start()], '(' + expression + ')'
    elif len(res) > 2:
        raise ValueError("The condition {} cannot be parsed".format(expression))
    else:
        # Then let's handle Greater Less and Equal
        regex_str = '|'.join([regex for regex in BOOLEAN_RELATIONALS])
        regexp = re.compile(regex_str)
        res = list(regexp.finditer(expression))

        if len(res) == 1:
            return expression[0:res[0].start()], '(' + expression + ')'

        if len(res) == 2:
            return expression[res[0].end():res[1].start()], '(' + expression + ')'

        if len(res) > 2:
            raise ValueError("The condition {} cannot be parsed".format(expression))

    return False


# symbol filter
def symbol_filter(expression, local_dict):
    """
    For a given regulatory expression it identifies the regulatory variable and parses it out as a sympy symbol.

    :param expression: str, the regulatory expression
    :param local_dict: dict, mapping between the variables in the expression and a Symbol Sympy's object
    :return: Symbol, the regulatory variable as a sympy Symbol object
    """

    sympy_expr = parse_expr(expression, evaluate=False, local_dict=local_dict)

    def get_symbol(atoms):

        for atom in atoms:

            if not atom.is_Atom:
                return get_symbol(atom.atoms())

            if atom.is_Atom and atom.is_Symbol:
                return atom

        return

    return get_symbol(sympy_expr.atoms())


# variable filter
def variable_from_str(expression, filter=True, special_chars=None, replaces=None):
    """
    For a given expression (string) it identifies the variable and parses it out as a sympy symbol.
    If more than one variable is present in the regulatory expression, a BaseException is raised

    It also applies the starts_with_digit_filter and special_chars_filter if filter is True. If the replaces
    dictionary is passed or reconstructed during filtering, the function can return the regulatory variable alias.
    Otherwise, the alias is equal to the regulatory variable ID depending on the filter condition.

    :param expression: str, expression
    :param filter: bool, if filter, special_chars_filter and starts_with_digit_filter are applied
    :param special_chars: dict, dict or None, special (or not) chars to be replaced. BOOL_SPECIAL_CHARS is used by default. special_chars argument must have unique keys and unique values!!
    :param replaces: dict, replaces occurred for aliases construction. None is the default.
    :return: regulatory variable Sympy Symbol, regulatory variable
    :return: alias: dict, the new variable name map to the old one

    """

    if filter:

        expression, new_replaces = special_chars_filter(expression, special_chars)
        expression, new_replaces_d = starts_with_digit_filter(expression)
        new_replaces.update(new_replaces_d)

    else:

        new_replaces = {}

    if replaces is None:
        replaces = {}

    replaces.update(new_replaces)

    sympify = parse_expr(expression, evaluate=False)
    aliases = {}

    if len(sympify.atoms()) > 1:
        raise BaseException("Something went wrong processing the variable {}".format(expression))

    if not isinstance(sympify, Symbol):
        raise BaseException("Something went wrong processing the variable {}".format(expression))

    old_reg_name = ''.join(sympify.name)

    for replace, special_char in replaces.items():
        old_reg_name = old_reg_name.replace(replace, special_char)

    aliases[sympify.name] = old_reg_name

    return sympify, aliases


# relational filter
def relational_filter(rule):
    """
    For a given boolean expression (string) finds all relational expressions and encloses all of them with
    parenthesis and replaces the ambiguous operators

    :param rule: str, the boolean expression
    :return: str, the new boolean expression having all relational enclosed by parenthesis
    """

    for relational in Boolean._SYMPY_RELATIONAL_REPLACE:
        if relational in rule:
            rule = rule.replace(relational, Boolean._SYMPY_RELATIONAL_REPLACE[relational])
        elif relational.upper() in rule:
            rule = rule.replace(relational.upper(), Boolean._SYMPY_RELATIONAL_REPLACE[relational])
        elif relational.title() in rule:
            rule = rule.replace(relational.title(), Boolean._SYMPY_RELATIONAL_REPLACE[relational])
        else:
            continue

    # regex to match a relational of any type: x>0; x<0; x>=0; x=>0;reverse order 0>x ...; with or without white spaces
    # ([a-zA-Z0-9_]+>[0-9]+)|([a-zA-Z0-9_]+\s>\s[0-9]+)

    regex_str = '|'.join(['([a-zA-Z0-9_]+' + regex + '[0-9.]+)' + '|' +
                          '([a-zA-Z0-9_]+\s' + regex + '\s[0-9.]+)' + '|' +
                          '([0-9.]+' + regex + '[a-zA-Z0-9_]+)' + '|' +
                          '([0-9.]+\s' + regex + '\s[a-zA-Z0-9_]+)'
                          for regex in BOOLEAN_EQUAL_RELATIONALS + BOOLEAN_RELATIONALS])

    regexp = re.compile(regex_str)
    res = list(regexp.finditer(rule))

    for match in set(map(lambda x: x.group(), res)):
        rule = re.sub(r'\b{}\b'.format(match), '(' + match + ')', rule)

    return rule


def bitwise_rule_filter(rule):
    """
    Transforms python boolean operators and, or and not to corresponding bitwise ones
    :param rule: str, boolean rule
    :return: str, new parsed boolean rule
    """

    for bool, bit_val in Boolean.SYMPY_REPLACE.items():

        rule = re.sub(r'\b{}\b'.format(bool), bit_val, rule)
        rule = re.sub(r'\b{}\b'.format(bool.upper()), bit_val, rule)
        rule = re.sub(r'\b{}\b'.format(bool.title()), bit_val, rule)

    return rule


# Boolean regulatory rule from str
def boolean_rule_from_str(rule):
    """
    It creates a new parsed regulatory rule, list all its elements (operators and operands) and creates a sympy
    expression.
    It also identifies all regulatory variables and returns them.

    The regulatory variables might be filtered and transformed with the special characters and starts with digit
    filters. The old regulatory variable ID is kept as an alias

    :param rule: str, boolean rule
    :returns rule: str, parsed boolean rule
    :return elements: list, all elements in the rule (operators and operands)
    :return sympify: Sympy Boolean object Or, And or Not, Sympy Symbol object, or Sympy Integral object
    :return symbols: dict, all regulatory variables in the regulatory rule. Keys stand for their new IDs and values stand for their sympy symbols
    :return symbols: dict, all regulatory variables in the regulatory rule. Keys stand for their new IDs and values stand for their aliases
    """

    regex_str = '|'.join(['(' + regex + '[0-9.]+)' + '|' +
                          '(' + regex + '\s[0-9.]+)' + '|'
                          for regex in BOOLEAN_EQUAL_RELATIONALS + BOOLEAN_RELATIONALS])
    regexp = re.compile(regex_str)

    rule, replaces = special_chars_filter(rule)
    rule, new_replaces = starts_with_digit_filter(rule)
    replaces.update(new_replaces)

    rule = bitwise_rule_filter(rule)
    rule = relational_filter(rule)

    elements = tokenize_infix_expression(rule)

    aliases = {}

    if len(elements) < 1:
        return rule, elements, Node(None), [], aliases, []

    if len(elements) == 1:

        if elements[0] == '1' or elements[0] == '0':
            return rule, elements, build_tree(exp=rule, rules=Boolean), [], [], []

    bool_tree = build_tree(exp=rule, rules=Boolean)

    operands = bool_tree.get_operands()
    conditionals = bool_tree.get_conditions()
    new_operands = []
    new_conditionals = []

    for op in operands:

        res = list(regexp.finditer(op))
        new_op = op

        for match in set(map(lambda x: x.group(), res)):
            new_op = re.sub(r'\b{}\b'.format(match), '', op)

        new_operands.append(new_op)
        if op in conditionals:
            new_conditionals.append(new_op)

        old_reg_name = ''.join(new_op)

        for replace, special_char in replaces.items():
            old_reg_name = old_reg_name.replace(replace, special_char)

        aliases[new_op] = [old_reg_name]

    return rule, elements, bool_tree, new_operands, aliases, new_conditionals



"""
##################################################################

GECKO model over COBRApy optimization test

##################################################################
"""
from geckopy.gecko import GeckoModel
from mewpy.simulation.cobra import GeckoSimulation
from mewpy.simulation import SimulationMethod
from mewpy.simulation.simulation import SimulationResult
from mewpy.problems.gecko import GeckoKOProblem, GeckoOUProblem
from mewpy.optimization.evaluation import BPCY, WYIELD, TargetFlux
from mewpy.optimization import EA
import mewpy.utils.utilities as utl
from collections import OrderedDict
from time import time

import warnings


ITERATIONS = 100


def gecko_ko(compound, display=False, filename=None):
    """ GECKO enzyme deletion example. 
    It runs a multi objective optimization for the increased production of a certain compound on yeast.
    The GECKO model is the yeast model companion from the GECKO paper "Improving the phenotype predictions
    of a yeast genome‐scale metabolic model by incorporating enzymatic constraints" https://doi.org/10.15252/msb.20167411.
    Runs over the GECKO original implementation. 
    
    :param compound: A target reaction identifier.
    :param display: Prints the best solution.
    :param filename: If given, saves the results as csv to filename.
         
    """

    warnings.filterwarnings("ignore")

    model = GeckoModel('single-pool', biomass_reaction_id='r_2111')
    model.solver = 'cplex'
    model.objective = {model.reactions.get_by_id('r_2111'): 1.0}
    envcond = OrderedDict()

    # the evaluation (objective) functions
    evaluator_1 = WYIELD("r_2111", compound)
    evaluator_2 = BPCY("r_2111", compound, uptake="r_1714_REV",
                       method=SimulationMethod.lMOMA)

    # The optimization problem
    # Notes:
    #  - A scale factor for the LP can be defined by setting the 'scalefactor' acordingly.
    #  - The scale factor is only used in the solver context and all results are scale free.
    problem = GeckoKOProblem(model,
                              fevaluation=[evaluator_1, evaluator_2],
                              envcond=envcond,
                              scalefactor=None,
                              candidate_max_size=30)

    # A new instance of the EA optimizer
    ea = EA(problem, max_generations=ITERATIONS, mp=True)
    # runs the optimization
    final_pop = ea.run()
    # optimization results
    if display:
        individual = max(final_pop)
        best = list(problem.decode(individual.candidate).keys())
        print('Best Solution: \n{0}'.format(str(best)))
    # save final population to file
    if filename:
        print("Simplifying and saving solutions to file")
        utl.population_to_csv(problem, final_pop, filename, simplify=False)


def gecko_ou(compound, display=False, filename=None):
    """ GECKO enzyme over/under expressoion example. 
    It runs a multi objective optimization for the increased production of a certain compound on yeast.
    The GECKO model is the yeast model companion from the GECKO paper "Improving the phenotype predictions
    of a yeast genome‐scale metabolic model by incorporating enzymatic constraints" https://doi.org/10.15252/msb.20167411. 
    Runs over the GECKO original implementation.
    
    :param compound: A target reaction identifier.
    :param display: Prints the best solution.
    :param filename: If given, saves the results as csv to filename.
         
    """

    model = GeckoModel('single-pool', biomass_reaction_id='r_2111')
    model.objective = 'r_2111'
    envcond = OrderedDict()

    # wild type reference values
    simulation = GeckoSimulation(model, envcond=envcond)
    res = simulation.simulate(method=SimulationMethod.pFBA)
    reference = res.fluxes

    # the evaluation (objective) functions
    evaluator_1 = WYIELD("r_2111", compound)
    evaluator_2 = BPCY("r_2111", compound, uptake="r_1714_REV",
                       method=SimulationMethod.lMOMA, reference=reference)

    # The optimization problem
    # Notes:
    #  - A scale factor for the LP can be defined by setting the 'scalefactor' acordingly.
    #  - The scale factor is only used in the solver context and all results are scale free.
    problem = GeckoOUProblem(model,
                              fevaluation=[evaluator_1, evaluator_2],
                              envcond=envcond,
                              reference=reference,
                              scalefactor=None,
                              candidate_max_size=30)

    # A new instance of the EA optimizer
    ea = EA(problem, max_generations=ITERATIONS, mp=False)
    # runs the optimization
    final_pop = ea.run()
    # optimization results
    if display:
        individual = max(final_pop)
        best = list(problem.decode(individual.candidate).keys())
        print('Best Solution: \n{0}'.format(str(best)))
    # save final population to file
    if filename:
        print("Simplifying and saving solutions to file")
        utl.population_to_csv(problem, final_pop, filename, simplify=False)


if __name__ == '__main__':

    N_EXP = 1

    compounds = {'SUC': 'r_2056',
                 'TYR': 'r_1913',
                 'PHE': 'r_1903',
                 'TRY': 'r_1912'}

    for k, v in compounds.items():
        for _ in range(N_EXP):
            millis = int(round(time() * 1000))
            gecko_ou(v, display=False,
                     filename="gecko_{}_OU_{}.csv".format(k, millis))

    for k, v in compounds.items():
        for _ in range(N_EXP):
            millis = int(round(time() * 1000))
            gecko_ko(v, display=False,
                     filename="gecko_{}_KO_{}.csv".format(k, millis))



from abc import abstractclassmethod
from collections import OrderedDict
from mewpy.simulation import SimulationMethod


class ModelContainer:
    """Interface for Model container.
       Provides an abstraction from models implementations.
    """

    @property
    def reactions(self):
        """
        :returns: A list of reaction identifiers.

        """
        raise NotImplementedError

    @property
    def genes(self):
        """
        :returns: A list of gene identifiers.

        """
        raise NotImplementedError

    @property
    def proteins(self):
        """
        :returns: A list of proteins identifiers.

        """
        raise NotImplementedError

    @property
    def metabolites(self):
        """
        :returns: A list of metabolite identifiers.

        """
        raise NotImplementedError

    @property
    def medium(self):
        raise NotImplementedError

    def get_drains(self):
        return NotImplementedError

    def get_gpr(self, reaction_id):
        """
        :returns: A string representation of the reaction GPR if exists None otherwise.
        """
        raise NotImplementedError

    def summary(self):
        print(f"Metabolites: {len(self.metabolites)}")
        print(f"Reactions: {len(self.reactions)}")
        print(f"Genes: {len(self.genes)}")

    def set_objective(self, reaction):
        raise NotImplementedError


class Simulator(ModelContainer):
    """
    Interface for simulators
    """

    @abstractclassmethod
    def simulate(self, objective=None, method=SimulationMethod.FBA, maximize=True, constraints=None, reference=None,
                 solver=None, **kwargs):
        """Abstract method to run a phenotype simulation.

        :returns: A SimulationResult.

        """
        raise NotImplementedError

    @abstractclassmethod
    def FVA(self, obj_frac=0, reactions=None, constraints=None, loopless=False, internal=None, solver=None):
        """ Abstract method to run Flux Variability Analysis (FVA).

        :returns: A dictionary of flux range values.

        """
        raise NotImplementedError


class SimulationResult(object):
    """Class that represents simulation results and performs operations over them."""

    def __init__(self, model, objective_value, fluxes=None, status=None, envcond=None, model_constraints=None,
                 simul_constraints=None, maximize=True):
        """
        :param model: A model instance.
        :param objective_value: The phenotype simulation objective value.
        :param dict fluxes: A dictionary of reaction fluxes values.
        :param status: The LP status.
        :param dict envcond: The environmental conditions of the phenotype simulation.
        :param dict model_constraints: Possible persistent additional constraints.
        :param dict simul_constraints: The simulation constraints.
        :param boolean maximize: Optimization direction.

        """
        self.model = model
        self.objective_value = objective_value
        self.status = status
        self.fluxes = fluxes
        self.maximize = maximize
        # Environmental conditions
        self.envcond = envcond if envcond else OrderedDict()
        # Persistent additional constraints not included in the environmental conditions
        self.model_constraints = model_constraints if model_constraints else OrderedDict()
        # Constraints specific to the simulation
        self.simulation_constraints = simul_constraints if simul_constraints else OrderedDict()

    def get_constraints(self):
        """
        :returns: All constraints applied during the simulation both persistent and simulation specific.
        """
        constraints = OrderedDict()
        constraints.update(self.envcond)
        constraints.update(self.model_constraints)
        constraints.update(self.simulation_constraints)
        return constraints

    def __repr__(self):
        return "objective: {}\nStatus: {}".format(self.objective_value, self.status)

    def __str__(self):
        return "objective: {}\nStatus: {}".format(self.objective_value, self.status)

    def get_net_conversion(self, biomassId=None):
        """Returns a string representation of the net conversion.

        :params str biosmassId: Biomass identifier (optional)
        :returns: A string representation of the net conversion.

        """

        left = ""
        right = ""
        firstLeft, firstRight = True, True

        ssFluxes = self.fluxes
        reactions = self.model.reactions
        for r_id in reactions.keys():
            fluxValue = ssFluxes[r_id]
            sub = reactions[r_id].get_substrates()
            prod = reactions[r_id].get_products()
            # if rId is a drain reaction
            if fluxValue != 0.0 and (len(sub) == 0 or len(prod) == 0):
                m = sub + prod
                if fluxValue < 0:
                    if firstLeft:
                        firstLeft = False
                    else:
                        left = left+" + "
                    left = left+str(-1*fluxValue)
                    left = left+" "+m[0]
                else:
                    if firstRight:
                        firstRight = False
                    else:
                        right = right+" + "
                    right = right+str(fluxValue)
                    right = right+" "+m[0]

        if biomassId and biomassId in ssFluxes.keys():
            biomassFlux = ssFluxes[biomassId]
            if biomassFlux > 0:
                if firstRight:
                    firstRight = False
                else:
                    right = right+" + "
                right = right + str(biomassFlux)
                right = right + " " + biomassId

        return left + " --> " + right



from inspyred.ec.emo import Pareto
from mewpy.visualization.plot import StreamingPlot
import numpy


def fitness_statistics(population):
    """Return the basic statistics of the population's fitness values.

    :param population: A population of individuals.

    """

    stats = {}

    population.sort(reverse=True)
    first = population[0].fitness

    if isinstance(first, Pareto):
        n = len(first.values)
        for i in range(n):
            f = [p.fitness.values[i] for p in population]
            worst_fit = min(f)
            best_fit = max(f)
            med_fit = numpy.median(f)
            avg_fit = numpy.mean(f)
            std_fit = numpy.std(f)
            stats['obj_{}'.format(i)] = {'best': best_fit, 'worst': worst_fit,
                                         'mean': avg_fit, 'median': med_fit, 'std': std_fit}
    else:
        worst_fit = population[-1].fitness
        best_fit = population[0].fitness
        f = [p.fitness for p in population]
        med_fit = numpy.median(f)
        avg_fit = numpy.mean(f)
        std_fit = numpy.std(f)
        stats['obj'] = {'best': best_fit, 'worst': worst_fit,
                        'mean': avg_fit, 'median': med_fit, 'std': std_fit}

    return stats


def results_observer(population, num_generations, num_evaluations, args):
    """
    Print the output of the evolutionary computation to a file with the follow fields:
    - number of generation
    - fitness of candidate
    - the solution candidates
    - the solution encoded candidates

    :param population: (list) the population of Individuals.
    :param num_generations: (int) the number of elapsed generations.
    :param num_evaluations: (int) the number of evaluations already performed.
    :param args: (dict) a dictionary of keyword arguments.

    """

    stats = fitness_statistics(population)
    title = "Gen    Eval|"
    values = "{0:>4} {1:>6}|".format(num_generations, num_evaluations)

    for key in stats:
        s = stats[key]
        title = title + "     Worst      Best    Median   Average   Std Dev|"
        values = values + "  {0:.6f}  {1:.6f}  {2:.6f}  {3:.6f}  {4:.6f}|".format(s['worst'],
                                                                                  s['best'],
                                                                                  s['median'],
                                                                                  s['mean'],
                                                                                  s['std'])
    if num_generations == 0:
        print(title)
    print(values)


class VisualizerObserver():

    def __init__(self, reference_front=None, reference_point=None, display_frequency=1, axis_labels=None,
                 non_dominated=True, print_stats=True):
        self.figure = None
        self.display_frequency = display_frequency
        self.reference_point = reference_point
        self.reference_front = reference_front
        self.print_stats = print_stats
        self.axis_labels = axis_labels
        self.non_dominated = non_dominated

    def update(self, population, num_generations, num_evaluations, args):
        generations = num_generations
        evaluations = num_evaluations

        if population:
            if self.non_dominated:
                pop = non_dominated_population(population)
            else:
                pop = population

            if self.figure is None:
                self.figure = StreamingPlot(axis_labels=self.axis_labels)
                solutions = []
                for i in range(len(pop)):
                    obj = pop[i].fitness
                    solutions.append(obj)
                self.figure.plot(solutions)

            if (generations % self.display_frequency) == 0:
                solutions = []
                for i in range(len(pop)):
                    obj = pop[i].fitness
                    solutions.append(obj)
                self.figure.update(solutions)
                self.figure.ax.set_title(
                    'Eval: {}'.format(evaluations), fontsize=13)

            if self.print_stats:
                results_observer(population, num_generations,
                                 num_evaluations, args)


def non_dominated_population(population, maximize=True):
    """ The non dominated solutions from the population.

    :param population: A list of individuals.
    :param maximize: (bool) Optimization direction.
    :returns: a list of non-dominated solutions.

    """
    population.sort(reverse=True)
    non_dominated = []
    for i in range(len(population)-1):
        individual = population[i]
        j = 0
        dominates = True
        while j < len(population) and dominates:
            if dominance_test(individual, population[j], maximize=maximize) == -1:
                dominates = False
            else:
                j += 1
        if dominates:
            non_dominated.append(individual)

    result = non_dominated
    return result


def dominance_test(solution1, solution2, maximize=True):
    """
    Testes Pareto dominance.

    :param solution1: The first solution.
    :param solution2: The second solution
    :param maximize: (bool) maximization (True) or minimization (False)

    :returns: 1, if the first solution dominates the second;
            -1, if the second solution dominates the first;
            0, if non of the solutions dominates the other.

    """
    best_is_one = 0
    best_is_two = 0

    if isinstance(solution1.fitness, Pareto):
        values1 = solution1.fitness.values
        values2 = solution2.fitness.values
    else:
        values1 = [solution1.fitness]
        values2 = [solution2.fitness]

    for i in range(len(values1)):
        value1 = values1[i]
        value2 = values2[i]
        if value1 != value2:
            if value1 < value2:
                best_is_two = 1
            if value1 > value2:
                best_is_one = 1

    if best_is_one > best_is_two:
        if maximize:
            result = 1
        else:
            result = -1
    elif best_is_two > best_is_one:
        if maximize:
            result = -1
        else:
            result = 1
    else:
        result = 0

    return result
